import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base.model import Model
from ..common.blocks import Conv2dReLU


def _upsample(x, size):
    return F.interpolate(x, size=size, mode='bilinear', align_corners=True)


class PyramidStage(nn.Module):

    def __init__(self, in_channels, out_channels, pool_size,
                 use_bathcnorm=True):
        super().__init__()
        if pool_size == 1:
            use_bathcnorm = False
        self.pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(pool_size, pool_size)),
            Conv2dReLU(in_channels, out_channels, (1, 1),
                       use_batchnorm=use_bathcnorm)
        )

    def forward(self, x):
        h, w = x.size(2), x.size(3)
        x = self.pool(x)
        x = _upsample(x, size=(h, w))
        return x


class PSPModule(nn.Module):
    def __init__(self, in_channels, sizes=(1, 2, 3, 6), use_bathcnorm=True):
        super().__init__()

        self.stages = nn.ModuleList([
            PyramidStage(in_channels, in_channels // len(sizes), size,
                         use_bathcnorm=use_bathcnorm) for size in sizes
        ])

    def forward(self, x):
        xs = [stage(x) for stage in self.stages] + [x]
        x = torch.cat(xs, dim=1)
        return x


class AUXModule(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features)

    def forward(self, x):
        x = F.adaptive_max_pool2d(x, output_size=(1, 1))
        x = x.view(-1, x.size(1))
        x = self.linear(x)
        return x


class PSPDecoder(Model):

    def __init__(
            self,
            encoder_channels,
            downsample_factor=8,
            use_batchnorm=True,
            psp_out_channels=512,
            final_channels=21,
            aux_output=False,
            dropout=0.2,
    ):
        super().__init__()
        self.downsample_factor = downsample_factor
        self.out_channels = self._get(encoder_channels)
        self.aux_output = aux_output
        self.dropout_factor = dropout

        self.psp = PSPModule(
            self.out_channels,
            sizes=(1, 2, 3, 6),
            use_bathcnorm=use_batchnorm,
        )

        self.conv = Conv2dReLU(
            self.out_channels * 2,
            psp_out_channels,
            kernel_size=1,
            use_batchnorm=use_batchnorm,
        )

        if self.dropout_factor:
            self.dropout = nn.Dropout2d(p=dropout)

        self.final_conv = nn.Conv2d(psp_out_channels, final_channels,
                                    kernel_size=(3, 3), padding=1)

        if self.aux_output:
            self.aux = AUXModule(self.out_channels, final_channels)

        self.initialize()

    def _get(self, xs):
        if self.downsample_factor == 4:
            return xs[3]
        elif self.downsample_factor == 8:
            return xs[2]
        elif self.downsample_factor == 16:
            return xs[1]
        else:
            raise ValueError(
                'Downsample factor should bi in [4, 8, 16], got {}'
                .format(self.downsample_factor))

    def forward(self, x):

        features = self._get(x)
        x = self.psp(features)
        x = self.conv(x)
        if self.dropout_factor:
            x = self.dropout(x)
        x = self.final_conv(x)
        x = F.interpolate(
            x,
            scale_factor=self.downsample_factor,
            mode='bilinear',
            align_corners=True
        )

        if self.training and self.aux_output:
            aux = self.aux(features)
            x = [x, aux]

        return x



import pickle
from typing import Tuple

import cv2
import numpy as np
from sklearn.metrics import confusion_matrix

from mlcomp.db.core import Session
from mlcomp.db.models import Report, ReportImg, Task, ReportTasks, ReportSeries
from mlcomp.db.providers import ReportProvider, \
    ReportLayoutProvider, \
    TaskProvider,\
    ReportImgProvider, \
    ReportTasksProvider, \
    ReportSeriesProvider, \
    DagProvider
from mlcomp.utils.img import resize_saving_ratio
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.utils.misc import now


class ClassificationReportBuilder:
    def __init__(
        self,
        session: Session,
        task: Task,
        layout: str,
        part: str = 'valid',
        name: str = 'img_classify',
        max_img_size: Tuple[int, int] = None,
        main_metric: str = 'accuracy',
        plot_count: int = 0
    ):
        self.session = session
        self.task = task
        self.layout = layout
        self.part = part
        self.name = name or 'img_classify'
        self.max_img_size = max_img_size
        self.main_metric = main_metric
        self.plot_count = plot_count

        self.dag_provider = DagProvider(session)
        self.report_provider = ReportProvider(session)
        self.layout_provider = ReportLayoutProvider(session)
        self.task_provider = TaskProvider(session)
        self.report_img_provider = ReportImgProvider(session)
        self.report_task_provider = ReportTasksProvider(session)
        self.report_series_provider = ReportSeriesProvider(session)

        self.project = self.task_provider.project(task.id).id
        self.layout = self.layout_provider.by_name(layout)
        self.layout_dict = yaml_load(self.layout.content)

    def create_base(self):
        report = Report(
            config=yaml_dump(self.layout_dict),
            time=now(),
            layout=self.layout.name,
            project=self.project,
            name=self.name
        )
        self.report_provider.add(report)
        self.report_task_provider.add(
            ReportTasks(report=report.id, task=self.task.id)
        )

        self.task.report = report.id
        self.task_provider.update()

    def process_scores(self, scores):
        for key, item in self.layout_dict['items'].items():
            item['name'] = key
            if item['type'] == 'series' and item['key'] in scores:
                series = ReportSeries(
                    name=item['name'],
                    value=float(scores[item['key']]),
                    epoch=0,
                    time=now(),
                    task=self.task.id,
                    part='valid',
                    stage='stage1'
                )

                self.report_series_provider.add(series)

    def process_pred(self, imgs: np.array, preds: np.array,
                     targets: np.array = None, attrs=None, scores=None):
        for key, item in self.layout_dict['items'].items():
            item['name'] = key
            if item['type'] != 'img_classify':
                continue

            report_imgs = []
            dag = self.dag_provider.by_id(self.task.dag)

            for i in range(len(imgs)):
                if self.plot_count <= 0:
                    break

                img = resize_saving_ratio(imgs[i], self.max_img_size)
                pred = preds[i]
                attr = attrs[i] if attrs else {}

                y = None
                score = None
                if targets is not None:
                    y = targets[i]
                    score = float(scores[self.main_metric][i])

                y_pred = pred.argmax()
                retval, buffer = cv2.imencode('.jpg', img)
                report_img = ReportImg(
                    group=item['name'],
                    epoch=0,
                    task=self.task.id,
                    img=buffer,
                    dag=self.task.dag,
                    part=self.part,
                    project=self.project,
                    y_pred=y_pred,
                    y=y,
                    score=score,
                    **attr
                )

                report_imgs.append(report_img)
                dag.img_size += report_img.size

            self.dag_provider.commit()
            self.report_img_provider.bulk_save_objects(report_imgs)

            if targets is not None and item.get('confusion_matrix'):
                matrix = confusion_matrix(
                    targets,
                    preds.argmax(axis=1),
                    labels=np.arange(preds.shape[1])
                )
                matrix = np.array(matrix)
                c = {'data': matrix}
                obj = ReportImg(
                    group=item['name'] + '_confusion',
                    epoch=0,
                    task=self.task.id,
                    img=pickle.dumps(c),
                    project=self.project,
                    dag=self.task.dag,
                    part=self.part
                )
                self.report_img_provider.add(obj)

            self.plot_count -= 1


__all__ = ['ClassificationReportBuilder']



from glob import glob
import os
from os.path import isdir, join
import hashlib
from typing import List, Tuple
import pkgutil
import sys
import pathspec
import pkg_resources
import pyclbr
import importlib

from sqlalchemy.orm import joinedload

from mlcomp import TASK_FOLDER, DATA_FOLDER, MODEL_FOLDER, INSTALL_DEPENDENCIES
from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType
from mlcomp.db.models import DagStorage, Dag, DagLibrary, File, Task
from mlcomp.utils.misc import now, to_snake
from mlcomp.db.providers import FileProvider, \
    DagStorageProvider, \
    TaskProvider, \
    DagLibraryProvider, DagProvider

from mlcomp.utils.config import Config
from mlcomp.utils.req import control_requirements, read_lines


def get_super_names(cls: pyclbr.Class):
    res = []
    if isinstance(cls.super, list):
        for s in cls.super:
            if isinstance(s, str):
                res.append(s)
            else:
                res.append(s.name)
                res.extend(get_super_names(s))
    elif isinstance(cls.super, pyclbr.Class):
        res.append(cls.super.name)
        res.extend(get_super_names(cls.super))
    elif isinstance(cls, str):
        res.append(cls)
    return res


class Storage:
    def __init__(self, session: Session, logger=None,
                 component: ComponentType = None,
                 max_file_size: int = 10 ** 5, max_count=10 ** 3):
        self.file_provider = FileProvider(session)
        self.provider = DagStorageProvider(session)
        self.task_provider = TaskProvider(session)
        self.library_provider = DagLibraryProvider(session)
        self.dag_provider = DagProvider(session)

        self.logger = logger
        self.component = component
        self.max_file_size = max_file_size
        self.max_count = max_count

    def log_info(self, message: str):
        if self.logger:
            self.logger.info(message, self.component)

    def copy_from(self, src: int, dag: Dag):
        storages = self.provider.query(DagStorage). \
            filter(DagStorage.dag == src). \
            all()
        libraries = self.library_provider.query(DagLibrary). \
            filter(DagLibrary.dag == src). \
            all()

        s_news = []
        for s in storages:
            s_new = DagStorage(
                dag=dag.id, file=s.file, path=s.path, is_dir=s.is_dir
            )
            s_news.append(s_new)
        l_news = []
        for l in libraries:
            l_new = DagLibrary(
                dag=dag.id, library=l.library, version=l.version
            )
            l_news.append(l_new)

        self.provider.add_all(s_news)
        self.library_provider.add_all(l_news)

    def _build_spec(self, folder: str):
        ignore_file = os.path.join(folder, 'file.ignore.txt')
        if not os.path.exists(ignore_file):
            ignore_patterns = []
        else:
            ignore_patterns = read_lines(ignore_file)
        ignore_patterns.extend(
            ['log', '/data', '/models', '__pycache__', '*.ipynb'])

        return pathspec.PathSpec.from_lines(
            pathspec.patterns.GitWildMatchPattern, ignore_patterns
        )

    def upload(self, folder: str, dag: Dag, control_reqs: bool = True):
        self.log_info('upload started')
        hashs = self.file_provider.hashs(dag.project)
        self.log_info('hashes are retrieved')

        all_files = []
        spec = self._build_spec(folder)

        files = glob(os.path.join(folder, '**'))
        for file in files[:]:
            path = os.path.relpath(file, folder)
            if spec.match_file(path) or path == '.':
                continue
            if os.path.isdir(file):
                child_files = glob(os.path.join(folder, file, '**'),
                                   recursive=True)
                files.extend(child_files)

        if self.max_count and len(files) > self.max_count:
            raise Exception(f'files count = {len(files)} '
                            f'But max count = {self.max_count}')

        self.log_info('list of files formed')

        folders_to_add = []
        files_to_add = []
        files_storage_to_add = []

        total_size_added = 0

        for o in files:
            path = os.path.relpath(o, folder)
            if spec.match_file(path) or path == '.':
                continue

            if isdir(o):
                folder_to_add = DagStorage(dag=dag.id, path=path, is_dir=True)
                folders_to_add.append(folder_to_add)
                continue
            content = open(o, 'rb').read()
            size = sys.getsizeof(content)
            if self.max_file_size and size > self.max_file_size:
                raise Exception(
                    f'file = {o} has size {size}.'
                    f' But max size is set to {self.max_file_size}')
            md5 = hashlib.md5(content).hexdigest()

            all_files.append(o)

            if md5 not in hashs:
                file = File(
                    md5=md5,
                    content=content,
                    project=dag.project,
                    dag=dag.id,
                    created=now()
                )
                hashs[md5] = file
                files_to_add.append(file)
                total_size_added += size

            file_storage = DagStorage(
                dag=dag.id, path=path, file=hashs[md5],
                is_dir=False)
            files_storage_to_add.append(file_storage)

        self.log_info('inserting DagStorage folders')

        if len(folders_to_add) > 0:
            self.provider.bulk_save_objects(folders_to_add)

        self.log_info('inserting Files')

        if len(files_to_add) > 0:
            self.file_provider.bulk_save_objects(files_to_add,
                                                 return_defaults=True)

        self.log_info('inserting DagStorage Files')

        if len(files_storage_to_add) > 0:
            for file_storage in files_storage_to_add:
                if isinstance(file_storage.file, File):
                    # noinspection PyUnresolvedReferences
                    file_storage.file = file_storage.file.id

            self.provider.bulk_save_objects(files_storage_to_add)

        dag.file_size += total_size_added

        self.dag_provider.update()

        if INSTALL_DEPENDENCIES and control_reqs:
            reqs = control_requirements(folder, files=all_files)
            for name, rel, version in reqs:
                self.library_provider.add(
                    DagLibrary(dag=dag.id, library=name, version=version)
                )

    def download_dag(self, dag: int, folder: str):
        os.makedirs(folder, exist_ok=True)

        items = self.provider.by_dag(dag)
        items = sorted(items, key=lambda x: x[1] is not None)
        for item, file in items:
            path = os.path.join(folder, item.path)
            if item.is_dir:
                os.makedirs(path, exist_ok=True)
            else:
                with open(path, 'wb') as f:
                    f.write(file.content)

    def download(self, task: int):
        task = self.task_provider.by_id(
            task, joinedload(Task.dag_rel, innerjoin=True)
        )
        folder = join(TASK_FOLDER, str(task.id))
        self.download_dag(task.dag, folder)

        config = Config.from_yaml(task.dag_rel.config)
        info = config['info']
        try:
            data_folder = os.path.join(DATA_FOLDER, info['project'])
            os.makedirs(data_folder, exist_ok=True)

            os.symlink(
                data_folder,
                os.path.join(folder, 'data'),
                target_is_directory=True
            )
        except FileExistsError:
            pass

        try:
            model_folder = os.path.join(MODEL_FOLDER, info['project'])
            os.makedirs(model_folder, exist_ok=True)

            os.symlink(
                model_folder,
                os.path.join(folder, 'models'),
                target_is_directory=True
            )
        except FileExistsError:
            pass

        sys.path.insert(0, folder)
        return folder

    def import_executor(
            self,
            folder: str,
            base_folder: str,
            executor: str,
            libraries: List[Tuple] = None
    ):

        sys.path.insert(0, base_folder)

        spec = self._build_spec(folder)
        was_installation = False

        folders = [
            p for p in glob(f'{folder}/*', recursive=True)
            if os.path.isdir(p) and not spec.match_file(p)
        ]
        folders += [folder]
        library_names = set(n for n, v in (libraries or []))
        library_versions = {n: v for n, v in (libraries or [])}

        for n in library_names:
            try:
                version = pkg_resources.get_distribution(n).version
                need_install = library_versions[n] != version
            except Exception:
                need_install = True

            if INSTALL_DEPENDENCIES and need_install:
                os.system(f'pip install {n}=={library_versions[n]}')
                was_installation = True

        def is_valid_class(cls: pyclbr.Class):
            return cls.name == executor or \
                   cls.name.lower() == executor or \
                   to_snake(cls.name) == executor

        def relative_name(path: str):
            rel = os.path.relpath(path, base_folder)
            parts = [str(p).split('.')[0] for p in rel.split(os.sep)]
            return '.'.join(parts)

        for (module_loader, module_name,
             ispkg) in pkgutil.iter_modules(folders):
            module = module_loader.find_module(module_name)
            rel_path = os.path.relpath(
                os.path.splitext(module.path)[0], base_folder
            ).replace('/', '.')
            try:
                classes = pyclbr.readmodule(rel_path, path=[base_folder])
            except Exception:
                continue
            for k, v in classes.items():
                if is_valid_class(v):
                    importlib.import_module(relative_name(module.path))
                    return True, was_installation

        return False, was_installation


__all__ = ['Storage']



import os
from io import BytesIO
from zipfile import ZipFile

import pandas as pd
import yaml

yaml.warnings({'YAMLLoadWarning': False})


def read_pandas(file):
    if file.endswith('.csv'):
        df = pd.read_csv(file)
    elif file.endswith('.parquet'):
        df = pd.read_parquet(file)
    else:
        raise Exception('Unknown file type')
    return df


def read_lines(file: str):
    return [l.strip() for l in open(file)]


def from_module_path(file: str, path: str):
    return os.path.join(os.path.dirname(file), path)


def yaml_load(text: str = None, file: str = None):
    stream = text or ''
    if file is not None:
        stream = open(file).read()
    res = yaml.load(stream, Loader=yaml.FullLoader)
    if res is None:
        return {}
    return res


def yaml_dump(data, file: str = None):
    res = yaml.dump(data, default_flow_style=False, sort_keys=False)
    if file:
        open(file, 'w').write(res)
    return res


def zip_folder(folder: str, dst: str = None):
    if dst is None:
        dst = BytesIO()

    with ZipFile(dst, 'w') as zip_obj:
        # Iterate over all the files in directory
        for folderName, subfolders, filenames in os.walk(folder):
            for filename in filenames:
                # create complete filepath of file in directory
                filePath = os.path.join(folderName, filename)
                # Add file to zip
                zip_obj.write(filePath, os.path.relpath(filePath, folder))
    return dst


__all__ = ['read_lines', 'from_module_path', 'yaml_load', 'yaml_dump',
           'zip_folder']



from os.path import basename, splitext

import pandas as pd

from sklearn.model_selection import GroupKFold


def file_group_kfold(
    n_splits: int,
    output: str,
    get_group=None,
    sort=False,
    must_equal=(),
    **files
):
    assert len(files) > 0, 'at lease 1 type of files is required'
    fold = GroupKFold(n_splits)
    keys = sorted(list(files))

    def get_name(file):
        return splitext(basename(file))[0]

    if sort:
        for k, v in files.items():
            files[k] = sorted(files[k], key=get_name)

    file_first = sorted(files[keys[0]])

    assert len(file_first) > n_splits, \
        f'at least {n_splits} files is required. Provided: {len(file_first)}'

    for k, v in files.items():
        assert len(files[k]) == len(file_first), \
            f'count of files in key = {k} is not the same as in {keys[0]}'

    for k, v in files.items():
        if k not in must_equal:
            continue
        for i in range(len(file_first)):
            names_equal = get_name(v[i]) == get_name(file_first[i])
            assert names_equal, \
                f'file name in {k} does not equal to {keys[0]}, ' \
                f'file name = {basename(v[i])}'

    df = pd.DataFrame(files)[keys]
    df['fold'] = 0

    groups = [
        i if not get_group else get_group(file)
        for i, file in enumerate(file_first)
    ]

    for i, (train_index,
            test_index) in enumerate(fold.split(groups, groups=groups)):
        df.loc[test_index, 'fold'] = i

    df = df.sample(frac=1)
    df.to_csv(output, index=False)



import numpy as np
import matplotlib.pyplot as plt

from sklearn.metrics import precision_recall_curve

from mlcomp.db.report_info.item import ReportLayoutItem
from mlcomp.utils.plot import figure_to_binary


class ReportLayoutPrecisionRecall(ReportLayoutItem):
    def plot(self, y: np.array, pred: np.array):
        p, r, t = precision_recall_curve(y, pred)
        fig, ax = plt.subplots(figsize=(4.2, 2.7))
        ax2 = ax.twinx()

        t = np.hstack([t, t[-1]])

        ax.plot(r, p)

        ax.set_xlabel('Recall')
        ax.set_ylabel('Precision')
        ax.set_ylim([0.0, 1.05])
        ax.set_xlim([0.0, 1.0])
        ax2.set_ylabel('Threashold')
        ax2.plot(r, t, c='red')
        return figure_to_binary(fig)


__all__ = ['ReportLayoutPrecisionRecall']



from enum import Enum
from mlcomp.utils.misc import to_snake


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    @classmethod
    def names(cls):
        return [e.name for e in cls]

    @classmethod
    def names_snake(cls):
        return [to_snake(n) for n in cls.names()]

    @classmethod
    def from_name(cls, name: str):
        if '_' in name or not name[0].isupper():
            return cls.names_snake().index(name)
        return cls.names().index(name)


class DagType(OrderedEnum):
    Standard = 0
    Pipe = 1


class TaskStatus(OrderedEnum):
    NotRan = 0
    Queued = 1
    InProgress = 2
    Failed = 3
    Stopped = 4
    Skipped = 5
    Success = 6


class TaskType(OrderedEnum):
    User = 0
    Train = 1
    Service = 2


class ComponentType(OrderedEnum):
    API = 0
    Supervisor = 1
    Worker = 2
    WorkerSupervisor = 3
    Client = 4


class LogStatus(OrderedEnum):
    Debug = 10
    Info = 20
    Warning = 30
    Error = 40



from collections import OrderedDict

import albumentations as A

from catalyst.dl import ConfigExperiment

from dataset import MnistDataset


class Experiment(ConfigExperiment):
    def get_datasets(self, stage: str, **kwargs):
        datasets = OrderedDict()

        train = MnistDataset(
            'data/train.csv',
            fold_csv='data/fold.csv',
            transforms=Experiment.get_test_transforms()
        )

        valid = MnistDataset(
            'data/train.csv',
            fold_csv='data/fold.csv',
            is_test=True,
            transforms=Experiment.get_test_transforms()
        )

        datasets['train'] = train
        datasets['valid'] = valid

        return datasets

    @staticmethod
    def get_test_transforms():
        return A.Compose([A.Normalize(mean=(0.485, ), std=(0.229, ))])



from collections import OrderedDict

import numpy as np

import torchvision
from torchvision import transforms

from catalyst.dl import ConfigExperiment


class Experiment(ConfigExperiment):
    @staticmethod
    def get_transforms(stage: str = None, mode: str = None):
        return torchvision.transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
            ]
        )

    def denormilize(self, img: np.array):
        # ``input[channel] = (input[channel] - mean[channel]) / std[channel]``

        res = np.zeros((img.shape[1], img.shape[2], 3), dtype=np.uint8)
        for i in range(res.shape[2]):
            res[:, :, i] = (img[i] * 0.5 + 0.5) * 255
        return res

    def get_datasets(self, stage: str, **kwargs):
        datasets = OrderedDict()

        trainset = torchvision.datasets.CIFAR10(
            root='./data',
            train=True,
            download=True,
            transform=Experiment.get_transforms(stage=stage, mode='train')
        )
        testset = torchvision.datasets.CIFAR10(
            root='./data',
            train=False,
            download=True,
            transform=Experiment.get_transforms(stage=stage, mode='valid')
        )

        datasets['train'] = trainset
        datasets['valid'] = testset

        return datasets



from collections import defaultdict

from sqlalchemy import and_

from mlcomp.db.enums import TaskStatus, TaskType
from mlcomp.db.models import TaskSynced, Task, Dag, Project, Computer
from mlcomp.db.providers.base import BaseDataProvider


class TaskSyncedProvider(BaseDataProvider):
    model = TaskSynced

    def for_computer(self, name: str):
        query = self.query(Task).filter(
            Task.status == TaskStatus.Success.value). \
            filter(Task.type <= TaskType.Train.value). \
            filter(Task.computer_assigned.__ne__(None)). \
            join(TaskSynced, and_(TaskSynced.task == Task.id,
                                  TaskSynced.computer == name), isouter=True).\
            filter(TaskSynced.task.__eq__(None))

        res = []
        by_computer_project = defaultdict(list)
        for task in query.all():
            if task.computer_assigned == name:
                continue
            dag = self.query(Dag).filter(Dag.id == task.dag).one()
            by_computer_project[(task.computer_assigned, dag.project)].append(
                task)

        for (c, p), tasks in by_computer_project.items():
            c = self.query(Computer).filter(Computer.name == c).one()
            p = self.query(Project).filter(Project.id == p).one()

            res.append((c, p, tasks))
        return res


__all__ = ['TaskSyncedProvider']



from sqlalchemy import func, case

from mlcomp.db.enums import LogStatus
from mlcomp.db.models import Step, Log
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.misc import to_snake, now


class StepProvider(BaseDataProvider):
    model = Step

    def _hierarchy(self, parent: dict, steps: list, start: int, end: int):
        current = start
        while current <= end:
            s = steps[current][0]
            if s.level == parent['level'] + 1:
                child = {**self.step_info(steps[current]), 'children': []}
                parent['children'].append(child)
                for i in range(current + 1, end + 1):
                    if steps[i][0].level <= s.level:
                        self._hierarchy(child, steps, current + 1, i - 1)
                        current = i
                        break
                    if i == len(steps) - 1:
                        self._hierarchy(child, steps, current + 1, i)
                else:
                    break
            else:
                current += 1

    def step_info(self, step):
        step, *log_status = step
        duration = ((step.finished if step.finished else now()) - step.started)
        res = {
            'id': step.id,
            'name': step.name,
            'level': step.level,
            'duration': duration.total_seconds(),
            'log_statuses': [
                {
                    'name': to_snake(e.name),
                    'count': s
                } for e, s in zip(LogStatus, log_status)
            ]
        }
        return res

    def get(self, task_id: int):
        log_status = []
        for s in LogStatus:
            log_status.append(
                func.sum(
                    case(
                        whens=[(Log.level == s.value, 1)],
                        else_=0
                    ).label(s.name)
                )
            )

        query = self.query(Step, *log_status).filter(Step.task == task_id
                                                     ).order_by(Step.id)
        query = query.join(Log, isouter=True).group_by(Step.id)
        steps = query.all()
        if len(steps) == 0:
            return []
        d = self.step_info(steps[0]) if len(steps) > 0 else dict()

        hierarchy = {**d, 'children': []}
        self._hierarchy(hierarchy, steps, 1, len(steps) - 1)
        return {'data': [hierarchy]}

    def last_for_task(self, id: int):
        return self.query(Step).filter(Step.task == id
                                       ).order_by(Step.started.desc()).first()

    def unfinished(self, task_id: int):
        return self.query(Step).filter(Step.task == task_id
                                       ).filter(Step.finished.__eq__(None)
                                                ).order_by(Step.started).all()


__all__ = ['StepProvider']



import numpy as np

from sklearn.metrics import classification_report

from mlcomp.db.report_info.item import ReportLayoutItem
from mlcomp.utils.plot import figure_to_binary, plot_classification_report


class ReportLayoutF1(ReportLayoutItem):
    def plot(self, y: np.array, pred: np.array):
        report = classification_report(y, pred)
        fig = plot_classification_report(report)
        return figure_to_binary(fig, dpi=70)


__all__ = ['ReportLayoutF1']



import functools
import torch.utils.model_zoo as model_zoo

from .resnet import resnet_encoders
from .dpn import dpn_encoders
from .vgg import vgg_encoders
from .senet import senet_encoders
from .densenet import densenet_encoders
from .inceptionresnetv2 import inception_encoders

from ._preprocessing import preprocess_input

encoders = {}
encoders.update(resnet_encoders)
encoders.update(dpn_encoders)
encoders.update(vgg_encoders)
encoders.update(senet_encoders)
encoders.update(densenet_encoders)
encoders.update(inception_encoders)


def get_encoder(name, encoder_weights=None):
    Encoder = encoders[name]['encoder']
    encoder = Encoder(**encoders[name]['params'])
    encoder.out_shapes = encoders[name]['out_shapes']

    if encoder_weights is not None:
        settings = encoders[name]['pretrained_settings'][encoder_weights]
        encoder.load_state_dict(model_zoo.load_url(settings['url']))

    return encoder


def get_encoder_names():
    return list(encoders.keys())


def get_preprocessing_params(encoder_name, pretrained='imagenet'):
    settings = encoders[encoder_name]['pretrained_settings']

    if pretrained not in settings.keys():
        raise ValueError(
            'Avaliable pretrained options {}'.format(settings.keys()))

    formatted_settings = {}
    formatted_settings['input_space'] = settings[pretrained].get('input_space')
    formatted_settings['input_range'] = settings[pretrained].get('input_range')
    formatted_settings['mean'] = settings[pretrained].get('mean')
    formatted_settings['std'] = settings[pretrained].get('std')
    return formatted_settings


def get_preprocessing_fn(encoder_name, pretrained='imagenet'):
    params = get_preprocessing_params(encoder_name, pretrained=pretrained)
    return functools.partial(preprocess_input, **params)



import albumentations as A
import numpy as np

from torch.utils.data import Dataset

from mlcomp.contrib.torch.tensors import flip


class TtaWrap(Dataset):
    def __init__(self, dataset: Dataset, tfms=()):
        self.dataset = dataset
        self.tfms = tfms

    def __getitem__(self, item):
        return self.dataset[item]

    def __len__(self):
        return len(self.dataset)

    def inverse(self, a: np.array):
        last_dim = len(a.shape) - 1
        for t in self.tfms:
            if isinstance(t, A.HorizontalFlip):
                a = flip(a, last_dim)
            elif isinstance(t, A.VerticalFlip):
                a = flip(a, last_dim - 1)
            elif isinstance(t, A.Transpose):
                axis = (0, 1, 3, 2) if len(a.shape) == 4 else (0, 2, 1)
                a = a.permute(*axis)

        return a


__all__ = ['TtaWrap']



import os
from abc import ABC, abstractmethod

from mlcomp.worker.executors import Executor
from mlcomp.worker.executors.base.equation import Equation


@Executor.register
class Infer(Equation, ABC):
    def __init__(
            self,
            test: bool = False,
            layout: str = None,
            plot_count: int = 0,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.test = test
        self.layout = layout
        self.plot_count = plot_count

    @abstractmethod
    def count(self):
        pass

    @abstractmethod
    def plot(self, preds):
        pass

    def key(self):
        return 'y'

    @abstractmethod
    def save(self, preds, folder: str):
        pass

    @abstractmethod
    def save_final(self, folder: str):
        pass

    @abstractmethod
    def create_base(self):
        pass

    @abstractmethod
    def adjust_part(self, part):
        pass

    def work(self):
        folder = 'data/pred'
        os.makedirs(folder, exist_ok=True)

        self.create_base()
        parts = self.generate_parts(self.count())

        for preds in self.solve(self.key(), parts):
            self.save(preds, folder)

            if self.layout:
                self.plot(preds)

        self.save_final(folder)



import re
import os
from os.path import join
from uuid import uuid4

import click
import pandas as pd

from mlcomp.contrib.scripts.split import file_group_kfold

current_folder = os.getcwd()


@click.group()
def main():
    pass


@main.command()
@click.argument('img_path')
@click.argument('n_splits', type=int)
@click.option('--group-regex')
def split_classify(img_path: str,
                   n_splits: int,
                   group_regex: str = None):
    output = join(current_folder, 'fold.csv')
    get_group = None
    if group_regex:
        pattern = re.compile(group_regex)

        def get_group(x):
            match = pattern.match(x)
            if not match:
                return str(uuid4())
            return match.group(1)

    images_labels = [(img, sub_folder)
                     for sub_folder in os.listdir(img_path)
                     for img in os.listdir(join(img_path, sub_folder))]

    file_group_kfold(n_splits, output, get_group=get_group,
                     image=[join(label, img) for img, label in images_labels],
                     label=[label for img, label in images_labels]
                     )


@main.command()
@click.argument('img_path')
@click.argument('mask_path')
@click.argument('n_splits', type=int)
@click.option('--group-regex')
def split_segment(img_path: str,
                  mask_path: str,
                  n_splits: int,
                  group_regex: str = None):
    output = join(current_folder, 'fold.csv')
    get_group = None
    if group_regex:
        pattern = re.compile(group_regex)

        def get_group(x):
            match = pattern.match(x)
            if not match:
                return str(uuid4())
            return match.group(1)

    file_group_kfold(n_splits, output, get_group=get_group,
                     image=os.listdir(img_path),
                     mask=os.listdir(mask_path),
                     sort=True,
                     must_equal=['image', 'mask']
                     )


@main.command()
@click.argument('img_path')
def split_test_img(img_path: str):
    output = join(current_folder, 'fold_test.csv')
    df = pd.DataFrame({
        'image': sorted(list(os.listdir(img_path))),
        'fold': 0
    })
    df.to_csv(output, index=False)


if __name__ == '__main__':
    main()



import datetime
import time
import traceback
from typing import List

from sqlalchemy.orm.exc import ObjectDeletedError

from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType, TaskStatus, TaskType
from mlcomp.db.models import Task, Auxiliary
from mlcomp.db.providers import \
    ComputerProvider, \
    TaskProvider, \
    DockerProvider, \
    AuxiliaryProvider, DagProvider, LogProvider
from mlcomp.utils.io import yaml_dump, yaml_load
from mlcomp.utils.logging import create_logger
from mlcomp.utils.misc import now
from mlcomp.worker.tasks import execute
from mlcomp.utils.schedule import start_schedule
import mlcomp.worker.tasks as celery_tasks


class SupervisorBuilder:
    def __init__(self):
        self.session = Session.create_session(key='SupervisorBuilder')
        self.logger = create_logger(self.session, 'SupervisorBuilder')
        self.provider = None
        self.computer_provider = None
        self.docker_provider = None
        self.auxiliary_provider = None
        self.dag_provider = None
        self.log_provider = None
        self.queues = None
        self.not_ran_tasks = None
        self.dep_status = None
        self.computers = None
        self.auxiliary = {}

        self.tasks = []
        self.tasks_stop = []
        self.dags_start = []
        self.sent_tasks = 0

    def create_base(self):
        self.session.commit()

        self.provider = TaskProvider(self.session)
        self.computer_provider = ComputerProvider(self.session)
        self.docker_provider = DockerProvider(self.session)
        self.auxiliary_provider = AuxiliaryProvider(self.session)
        self.dag_provider = DagProvider(self.session)
        self.log_provider = LogProvider(self.session)

        self.queues = [
            f'{d.computer}_{d.name}' for d in self.docker_provider.all()
            if d.last_activity >= now() - datetime.timedelta(seconds=15)
        ]

        self.auxiliary['queues'] = self.queues

    def load_tasks(self):
        self.tasks = self.provider.by_status(TaskStatus.NotRan,
                                             TaskStatus.InProgress,
                                             TaskStatus.Queued)

        not_ran_tasks = [t for t in self.tasks if
                         t.status == TaskStatus.NotRan.value]

        self.not_ran_tasks = [task for task in not_ran_tasks if not task.debug]
        self.not_ran_tasks = sorted(
            self.not_ran_tasks, key=lambda x: x.gpu or 0,
            reverse=True)

        self.logger.debug(
            f'Found {len(not_ran_tasks)} not ran tasks',
            ComponentType.Supervisor
        )

        self.dep_status = self.provider.dependency_status(self.not_ran_tasks)

        self.auxiliary['not_ran_tasks'] = [
            {
                'id': t.id,
                'name': t.name,
                'dep_status': [
                    TaskStatus(s).name
                    for s in self.dep_status.get(t.id, set())
                ]
            } for t in not_ran_tasks[:5]
        ]

    def load_computers(self):
        computers = self.computer_provider.computers()
        for computer in computers.values():
            computer['gpu'] = [0] * computer['gpu']
            computer['ports'] = set()
            computer['cpu_total'] = computer['cpu']
            computer['memory_total'] = computer['memory']
            computer['gpu_total'] = len(computer['gpu'])
            computer['can_process_tasks'] = computer['can_process_tasks']

        tasks = [
            t for t in self.tasks if
            t.status in [TaskStatus.InProgress.value,
                         TaskStatus.Queued.value]
        ]

        for task in tasks:
            if task.computer_assigned is None:
                continue
            assigned = task.computer_assigned
            comp_assigned = computers[assigned]
            comp_assigned['cpu'] -= task.cpu

            if task.gpu_assigned is not None:
                for g in task.gpu_assigned.split(','):
                    comp_assigned['gpu'][int(g)] = task.id
            comp_assigned['memory'] -= task.memory * 1024

            info = yaml_load(task.additional_info)
            if 'distr_info' in info:
                dist_info = info['distr_info']
                if dist_info['rank'] == 0:
                    comp_assigned['ports'].add(dist_info['master_port'])

        self.computers = [
            {
                **value, 'name': name
            } for name, value in computers.items()
        ]

        self.auxiliary['computers'] = self.computers

    def process_to_celery(self, task: Task, queue: str, computer: dict):
        r = execute.apply_async((task.id,), queue=queue, retry=False)
        task.status = TaskStatus.Queued.value
        task.computer_assigned = computer['name']
        task.docker_assigned = queue.split('_')[1]
        task.celery_id = r.id

        if task.computer_assigned is not None:
            if task.gpu_assigned:
                for g in map(int, task.gpu_assigned.split(',')):
                    computer['gpu'][g] = task.id
            computer['cpu'] -= task.cpu
            computer['memory'] -= task.memory * 1024

        self.logger.info(
            f'Sent task={task.id} to celery. Queue = {queue} '
            f'Task status = {task.status} Celery_id = {r.id}',
            ComponentType.Supervisor)
        self.provider.update()

    def create_service_task(
            self,
            task: Task,
            gpu_assigned=None,
            distr_info: dict = None,
            resume: dict = None
    ):
        new_task = Task(
            name=task.name,
            computer=task.computer,
            executor=task.executor,
            status=TaskStatus.NotRan.value,
            type=TaskType.Service.value,
            gpu_assigned=gpu_assigned,
            parent=task.id,
            report=task.report,
            dag=task.dag
        )
        new_task.additional_info = task.additional_info

        if distr_info:
            additional_info = yaml_load(new_task.additional_info)
            additional_info['distr_info'] = distr_info
            new_task.additional_info = yaml_dump(additional_info)

        if resume:
            additional_info = yaml_load(new_task.additional_info)
            additional_info['resume'] = resume
            new_task.additional_info = yaml_dump(additional_info)

        return self.provider.add(new_task)

    def find_port(self, c: dict, docker_name: str):
        docker = self.docker_provider.get(c['name'], docker_name)
        ports = list(map(int, docker.ports.split('-')))
        for p in range(ports[0], ports[1] + 1):
            if p not in c['ports']:
                return p
        raise Exception(f'All ports in {c["name"]} are taken')

    def _process_task_valid_computer(self, task: Task, c: dict,
                                     single_node: bool):
        if not c['can_process_tasks']:
            return 'this computer can not process tasks'

        if task.computer is not None and task.computer != c['name']:
            return 'name set in the config!= name of this computer'

        if task.cpu > c['cpu']:
            return f'task cpu = {task.cpu} > computer' \
                   f' free cpu = {c["cpu"]}'

        if task.memory > c['memory']:
            return f'task cpu = {task.cpu} > computer ' \
                   f'free memory = {c["memory"]}'

        queue = f'{c["name"]}_' \
                f'{task.dag_rel.docker_img or "default"}'
        if queue not in self.queues:
            return f'required queue = {queue} not in queues'

        if task.gpu > 0 and not any(g == 0 for g in c['gpu']):
            return f'task requires gpu, but there is not any free'

        free_gpu = sum(g == 0 for g in c['gpu'])
        if single_node and task.gpu > free_gpu:
            return f'task requires {task.gpu} ' \
                   f'but there are only {free_gpu} free'

    def _process_task_get_computers(
            self, executor: dict, task: Task, auxiliary: dict
    ):
        single_node = executor.get('single_node', True)

        computers = []
        for c in self.computers:
            error = self._process_task_valid_computer(task, c, single_node)
            auxiliary['computers'].append({'name': c['name'], 'error': error})
            if not error:
                computers.append(c)

        if task.gpu > 0 and single_node and len(computers) > 0:
            computers = sorted(
                computers,
                key=lambda x: sum(g == 0 for g in c['gpu']),
                reverse=True
            )[:1]

        free_gpu = sum(sum(g == 0 for g in c['gpu']) for c in computers)
        if task.gpu > free_gpu:
            auxiliary['not_valid'] = f'gpu required by the ' \
                                     f'task = {task.gpu},' \
                                     f' but there are only {free_gpu} ' \
                                     f'free gpus'
            return []
        return computers

    def _process_task_to_send(
            self, executor: dict, task: Task, computers: List[dict]
    ):
        distr = executor.get('distr', True)
        to_send = []
        for computer in computers:
            queue = f'{computer["name"]}_' \
                    f'{task.dag_rel.docker_img or "default"}'

            if task.gpu_max > 1 and distr:
                for index, task_taken_gpu in enumerate(computer['gpu']):
                    if task_taken_gpu:
                        continue
                    to_send.append([computer, queue, index])

                    if len(to_send) >= task.gpu_max:
                        break

                if len(to_send) >= task.gpu_max:
                    break
            elif task.gpu_max > 0:
                cuda_devices = []
                for index, task_taken_gpu in enumerate(computer['gpu']):
                    if task_taken_gpu:
                        continue

                    cuda_devices.append(index)
                    if len(cuda_devices) >= task.gpu_max:
                        break

                task.gpu_assigned = ','.join(map(str, cuda_devices))
                self.provider.commit()

                self.process_to_celery(task, queue, computer)
            else:
                self.process_to_celery(task, queue, computer)
                break
        return to_send

    def process_task(self, task: Task):
        auxiliary = self.auxiliary['process_tasks'][-1]
        auxiliary['computers'] = []

        config = yaml_load(task.dag_rel.config)
        executor = config['executors'][task.executor]

        computers = self._process_task_get_computers(executor, task, auxiliary)
        if len(computers) == 0:
            return

        to_send = self._process_task_to_send(executor, task, computers)
        auxiliary['to_send'] = to_send[:5]
        additional_info = yaml_load(task.additional_info)

        rank = 0
        master_port = None
        if len(to_send) > 0:

            master_port = self.find_port(
                to_send[0][0], to_send[0][1].split('_')[1]
            )
            computer_names = {c['name'] for c, _, __ in to_send}
            if len(computer_names) == 1:
                task.computer_assigned = list(computer_names)[0]

        for computer, queue, gpu_assigned in to_send:
            main_cmp = to_send[0][0]
            # noinspection PyTypeChecker
            ip = 'localhost' if computer['name'] == main_cmp['name'] \
                else main_cmp['ip']

            distr_info = {
                'master_addr': ip,
                'rank': rank,
                'local_rank': gpu_assigned,
                'master_port': master_port,
                'world_size': len(to_send),
                'master_computer': main_cmp['name']
            }
            service_task = self.create_service_task(
                task,
                distr_info=distr_info,
                gpu_assigned=gpu_assigned,
                resume=additional_info.get('resume')
            )
            self.process_to_celery(service_task, queue, computer)
            rank += 1
            main_cmp['ports'].add(master_port)

        if len(to_send) > 0:
            task.status = TaskStatus.Queued.value
            self.sent_tasks += len(to_send)

    def process_tasks(self):
        self.auxiliary['process_tasks'] = []

        for task in self.not_ran_tasks:
            auxiliary = {'id': task.id, 'name': task.name}
            self.auxiliary['process_tasks'].append(auxiliary)

            if task.dag_rel is None:
                task.dag_rel = self.dag_provider.by_id(task.dag)

            if TaskStatus.Stopped.value in self.dep_status[task.id] \
                    or TaskStatus.Failed.value in self.dep_status[task.id] or \
                    TaskStatus.Skipped.value in self.dep_status[task.id]:
                auxiliary['not_valid'] = 'stopped or failed in dep_status'
                self.provider.change_status(task, TaskStatus.Skipped)
                continue

            if len(self.dep_status[task.id]) != 0 \
                    and self.dep_status[task.id] != {TaskStatus.Success.value}:
                auxiliary['not_valid'] = 'not all dep tasks are finished'
                continue
            self.process_task(task)

        self.auxiliary['process_tasks'] = self.auxiliary['process_tasks'][:5]

    def _stop_child_tasks(self, task: Task):
        self.provider.commit()

        children = self.provider.children(task.id, [Task.dag_rel])
        dags = [c.dag_rel for c in children]
        for c, d in zip(children, dags):
            celery_tasks.stop(self.logger, self.session, c, d)

    def _correct_catalyst_hangs(self, task, statuses):
        if task.type != TaskType.Train.value:
            return
        success = statuses[TaskStatus.Success]
        in_progress = statuses[TaskStatus.InProgress]

        sum_statuses = sum(statuses.values())
        if (success + in_progress == sum_statuses) \
                and in_progress > 0 and success > 0:
            child_tasks = self.provider.children(task.id)
            for t in child_tasks:
                if t.status == TaskStatus.InProgress.value:
                    response = celery_tasks.kill.apply_async(
                        (t.pid,),
                        queue=f'{t.computer_assigned}_{t.docker_assigned}',
                        retry=False)

                    result = response.get()
                    if result:
                        t.status = TaskStatus.Success.value
            self.provider.commit()

    def _correct_catalyst_fails(self, task):
        if task.type != TaskType.Train.value:
            return

        messages = [
            'CUDA error: an illegal memory access was encountered',
            'CUDA error: device-side assert triggered',
            'CUDNN_STATUS_INTERNAL_ERROR'
        ]
        children = self.provider.children(task.id, [Task.dag_rel])
        children = sorted(children, key=lambda x: x.id, reverse=True)
        for c in children:
            if c.status == TaskStatus.Failed.value:
                logs = self.log_provider.last(task=c.id, count=1)
                if len(logs) > 0:
                    message = logs[0][0].message
                    for m in messages:
                        if m in message:
                            self.logger.info(
                                'Restart DAG: correct_catalyst_fails',
                                ComponentType.Supervisor,
                                None,
                                task.id
                            )
                            self.start_dag(task.dag)
                            return

    def process_parent_tasks(self):
        tasks = self.provider.parent_tasks_stats()

        was_change = False
        for task, started, finished, statuses in tasks:
            self._correct_catalyst_hangs(task, statuses)

            status = task.status
            if statuses[TaskStatus.Failed] > 0:
                status = TaskStatus.Failed.value
            elif statuses[TaskStatus.Skipped] > 0:
                status = TaskStatus.Skipped.value
            elif statuses[TaskStatus.Queued] > 0:
                status = TaskStatus.Queued.value
            elif statuses[TaskStatus.InProgress] > 0:
                status = TaskStatus.InProgress.value
            elif statuses[TaskStatus.Success] > 0:
                status = TaskStatus.Success.value

            if status != task.status:
                if status == TaskStatus.InProgress.value:
                    task.started = started
                elif status >= TaskStatus.Failed.value:
                    task.started = started
                    task.finished = finished

                    self._stop_child_tasks(task)
                    self._correct_catalyst_fails(task)

                was_change = True
                task.status = status

        if was_change:
            self.provider.commit()

        self.auxiliary['parent_tasks_stats'] = [
            {
                'name': task.name,
                'id': task.id,
                'started': task.started,
                'finished': finished,
                'statuses': [
                    {
                        'name': k.name,
                        'count': v
                    } for k, v in statuses.items()
                ],
            } for task, started, finished, statuses in tasks[:5]
        ]

    def write_auxiliary(self):
        self.auxiliary['duration'] = (now() - self.auxiliary['time']). \
            total_seconds()

        auxiliary = Auxiliary(
            name='supervisor', data=yaml_dump(self.auxiliary)
        )
        if len(auxiliary.data) > 16000:
            return

        self.auxiliary_provider.create_or_update(auxiliary, 'name')

    def stop_tasks(self, tasks: List[Task]):
        self.tasks_stop.extend([t.id for t in tasks])

    def process_stop_tasks(self):
        # Stop not running tasks
        if len(self.tasks_stop) == 0:
            return

        tasks = self.provider.by_ids(self.tasks_stop)
        tasks_not_ran = [t.id for t in tasks if
                         t.status in [TaskStatus.NotRan.value,
                                      TaskStatus.Queued.value]]
        tasks_started = [t for t in tasks if
                         t.status in [TaskStatus.InProgress.value]]
        tasks_started_ids = [t.id for t in tasks_started]

        self.provider.change_status_all(tasks=tasks_not_ran,
                                        status=TaskStatus.Skipped)

        pids = []
        for task in tasks_started:
            if task.pid:
                pids.append((task.computer_assigned, task.pid))

            additional_info = yaml_load(task.additional_info)
            for p in additional_info.get('child_processes', []):
                pids.append((task.computer_assigned, p))

        for computer, queue in self.docker_provider.queues_online():
            pids_computer = [p for c, p in pids if c == computer]
            if len(pids_computer) > 0:
                celery_tasks.kill_all.apply_async((pids_computer,),
                                                  queue=queue,
                                                  retry=False)

        self.provider.change_status_all(tasks=tasks_started_ids,
                                        status=TaskStatus.Stopped)

        self.tasks_stop = []

    def fast_check(self):
        if self.provider is None or self.computer_provider is None:
            return False

        if self.not_ran_tasks is None or self.queues is None:
            return False

        if len(self.tasks_stop) > 0:
            return False

        if len(self.dags_start) > 0:
            return False

        if len(self.auxiliary.get('to_send', [])) > 0:
            return False

        queues = set([
            f'{d.computer}_{d.name}' for d in self.docker_provider.all()
            if d.last_activity >= now() - datetime.timedelta(seconds=15)
        ])

        queues_set = set(queues)
        queues_set2 = set(self.queues)

        if queues_set != queues_set2:
            return False

        tasks = self.provider.by_status(TaskStatus.NotRan,
                                        TaskStatus.Queued,
                                        TaskStatus.InProgress)
        tasks_set = {t.id for t in tasks if
                     t.status == TaskStatus.NotRan.value and not t.debug}
        tasks_set2 = {t.id for t in self.tasks if
                      t.status == TaskStatus.NotRan.value}

        if tasks_set != tasks_set2:
            return False

        tasks_set = {t.id for t in tasks if
                     t.status == TaskStatus.InProgress.value}
        tasks_set2 = {t.id for t in self.tasks if
                      t.status == TaskStatus.InProgress.value}

        if tasks_set != tasks_set2:
            return False

        tasks_set = {t.id for t in tasks if
                     t.status == TaskStatus.Queued.value}
        tasks_set2 = {t.id for t in self.tasks if
                      t.status == TaskStatus.Queued.value}

        if tasks_set != tasks_set2:
            return False

        return True

    def start_dag(self, id: int):
        self.dags_start.append(id)

    def process_start_dags(self):
        if len(self.dags_start) == 0:
            return

        for id in self.dags_start:
            can_start_statuses = [
                TaskStatus.Failed.value, TaskStatus.Skipped.value,
                TaskStatus.Stopped.value
            ]

            tasks = self.provider.by_dag(id)
            children_all = self.provider.children([t.id for t in tasks])

            def find_resume(task):
                children = [c for c in children_all if c.parent == task.id]
                children = sorted(children, key=lambda x: x.id, reverse=True)

                if len(children) > 0:
                    for c in children:
                        if c.parent != task.id:
                            continue

                        info = yaml_load(c.additional_info)
                        if 'distr_info' not in info:
                            continue

                        if info['distr_info']['rank'] == 0:
                            return {
                                'master_computer': c.computer_assigned,
                                'master_task_id': c.id,
                                'load_last': True
                            }
                    raise Exception('Master task not found')
                else:
                    return {
                        'master_computer': task.computer_assigned,
                        'master_task_id': task.id,
                        'load_last': True
                    }

            for t in tasks:
                if t.parent:
                    t.continued = True
                    continue

                if t.status not in can_start_statuses:
                    continue

                if t.type == TaskType.Train.value:
                    info = yaml_load(t.additional_info)
                    info['resume'] = find_resume(t)
                    t.additional_info = yaml_dump(info)

                t.status = TaskStatus.NotRan.value
                t.pid = None
                t.started = None
                t.finished = None
                t.computer_assigned = None
                t.celery_id = None
                t.worker_index = None
                t.docker_assigned = None

        self.provider.commit()
        self.dags_start = []

    def build(self):
        try:
            # if self.fast_check():
            #     return

            self.auxiliary = {'time': now()}

            self.create_base()

            self.process_stop_tasks()

            self.process_start_dags()

            self.process_parent_tasks()

            self.load_tasks()

            self.load_computers()

            self.process_tasks()

            self.write_auxiliary()

        except ObjectDeletedError:
            pass
        except Exception as e:
            if Session.sqlalchemy_error(e):
                Session.cleanup(key='SupervisorBuilder')
                self.session = Session.create_session(key='SupervisorBuilder')
                self.logger = create_logger(self.session, 'SupervisorBuilder')

            self.logger.error(traceback.format_exc(), ComponentType.Supervisor)


def register_supervisor():
    builder = SupervisorBuilder()
    start_schedule([(builder.build, 1)])
    return builder


__all__ = ['SupervisorBuilder', 'register_supervisor']



from copy import deepcopy

import numpy as np
import cv2

from mlcomp.worker.executors import Executor
from mlcomp.worker.executors.valid import Valid
from mlcomp.worker.reports.classification import ClassificationReportBuilder

from dataset import MnistDataset
from experiment import Experiment


@Executor.register
class ValidMnist(Valid):
    def __init__(self, **kwargs):
        cache_names = ['y']
        super().__init__(cache_names=cache_names, layout='img_classify',
                         **kwargs)

        self.x_source = MnistDataset(
            file='data/train.csv',
            fold_csv='data/fold.csv',
            is_test=True,
            max_count=self.max_count,
            transforms=Experiment.get_test_transforms()
        )
        self.builder = None
        self.x = None
        self.scores = []

    def create_base(self):
        self.builder = ClassificationReportBuilder(
            session=self.session,
            task=self.task,
            layout=self.layout,
            name=self.name,
            plot_count=self.plot_count
        )
        self.builder.create_base()

    def count(self):
        return len(self.x_source)

    def score(self, preds):
        # noinspection PyUnresolvedReferences
        res = (preds.argmax(axis=1) == self.x.y).astype(np.float)
        self.scores.extend(res)
        return res

    def score_final(self):
        return np.mean(self.scores)

    def adjust_part(self, part):
        self.x = deepcopy(self.x_source)
        self.x.x = self.x.x[part[0]:part[1]]
        self.x.y = self.x.y[part[0]:part[1]]

    def plot(self, preds, scores):
        imgs = [
            cv2.cvtColor(
                ((row['features'][0] * 0.229 + 0.485) * 255).astype(np.uint8),
                cv2.COLOR_GRAY2BGR
            ) for row in self.x
        ]

        attrs = [
            {
                'attr1': p.argmax(),
                'attr2': t
            } for p, t in zip(preds, self.x.y)
        ]

        self.builder.process_pred(
            imgs=imgs,
            preds=preds,
            targets=self.x.y,
            attrs=attrs,
            scores={'accuracy': scores}
        )

    def plot_final(self, score):
        self.builder.process_scores({'accuracy': score})



import datetime
from typing import List, Union

from sqlalchemy import func, case
from sqlalchemy.orm import joinedload, aliased

from mlcomp.db.core import PaginatorOptions
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.db.enums import TaskType, DagType, TaskStatus
from mlcomp.utils.misc import to_snake, duration_format, now, parse_time
from mlcomp.db.models import Task, Project, Dag, TaskDependence, ReportTasks


class TaskProvider(BaseDataProvider):
    model = Task

    def _get_filter(self, query, filter: dict):
        if filter.get('dag'):
            query = query.filter(Task.dag == filter['dag'])

        if filter.get('name'):
            query = query.filter(Task.name.like(f'%{filter["name"]}%'))

        if filter.get('status'):
            status = [
                TaskStatus.from_name(k) for k, v in filter['status'].items()
                if v
            ]
            if len(status) > 0:
                query = query.filter(Task.status.in_(status))

        if filter.get('id_min'):
            query = query.filter(Task.id >= filter['id_min'])

        if filter.get('id_max'):
            query = query.filter(Task.id <= filter['id_max'])

        if filter.get('id'):
            query = query.filter(Task.id == filter['id'])

        if filter.get('project'):
            query = query.filter(Dag.project == filter['project'])

        if filter.get('created_min'):
            created_min = parse_time(filter['created_min'])
            query = query.filter(Dag.created >= created_min)
        if filter.get('created_max'):
            created_max = parse_time(filter['created_max'])
            query = query.filter(Dag.created <= created_max)
        if filter.get('last_activity_min'):
            last_activity_min = parse_time(filter['last_activity_min'])
            query = query.filter(Task.last_activity >= last_activity_min)
        if filter.get('last_activity_max'):
            last_activity_max = parse_time(filter['last_activity_max'])
            query = query.filter(Task.last_activity <= last_activity_max)
        if filter.get('report'):
            query = query.filter(Task.report is not None)
        if filter.get('parent'):
            query = query.filter(Task.parent == filter['parent'])

        types = filter.get('type', ['User', 'Train'])
        types = [TaskType.from_name(t) for t in types]
        query = query.filter(Task.type.in_(types))

        return query

    def get(self, filter: dict, options: PaginatorOptions):
        query = self.query(Task, Dag, Project.name). \
            join(Dag, Dag.id == Task.dag). \
            join(Project, Project.id == Dag.project)

        query = self._get_filter(query, filter)

        total = query.count()
        paginator = self.paginator(query, options)
        res = []

        for p, d, project_name in paginator.all():
            # noinspection PyDictCreation
            item = {**self.to_dict(p, rules=('-additional_info',))}
            item['status'] = to_snake(TaskStatus(item['status']).name)
            item['type'] = to_snake(TaskType(item['type']).name)
            item['dag_rel'] = self.to_dict(d)
            item['dag_rel']['project'] = {
                'id': item['dag_rel']['project'],
                'name': project_name
            }
            if p.started is None:
                delta = 0
            elif p.status == TaskStatus.InProgress.value:
                delta = (now() - p.started).total_seconds()
            else:
                finish = (p.finished or p.last_activity)
                delta = (finish - p.started).total_seconds()
            item['duration'] = duration_format(delta)
            res.append(item)

        if filter.get('report'):
            tasks_within_report = self.query(
                ReportTasks.task
            ).filter(ReportTasks.report == int(filter['report']))
            tasks_within_report = {t[0] for t in tasks_within_report}
            for r in res:
                r['report_full'] = r['id'] in tasks_within_report

        projects = self.query(Project.name, Project.id). \
            order_by(Project.id.desc()). \
            limit(20). \
            all()

        dags = self.query(Dag.name, Dag.id). \
            order_by(Dag.id.desc()). \
            limit(20). \
            all()

        projects = [{'name': name, 'id': id} for name, id in projects]
        dags = [{'name': name, 'id': id} for name, id in dags]

        dags_model = self.query(Dag.name, Dag.id, Dag.project). \
            filter(Dag.type == DagType.Pipe.value). \
            order_by(Dag.id.desc()). \
            all()

        dags_model_dict = []
        used_dag_names = set()

        for name, id, project in dags_model:
            if name in used_dag_names:
                continue

            dag = {'name': name, 'id': id, 'project': project}
            dags_model_dict.append(dag)
            used_dag_names.add(name)

        return {
            'total': total,
            'data': res,
            'projects': projects,
            'dags': dags,
            'dags_model': dags_model_dict
        }

    def last_tasks(self, min_time: datetime, status: int, joined_load=None):
        res = self.query(Task).filter(
            Task.finished >= min_time). \
            filter(Task.status == status)

        if joined_load is not None:
            for n in joined_load:
                res = res.options(joinedload(n, innerjoin=True))

        return res.all()

    def add_dependency(self, task_id: int, depend_id: int) -> None:
        self.add(TaskDependence(task_id=task_id, depend_id=depend_id))

    def by_id(self, id, options=None) -> Task:
        query = self.query(Task).filter(Task.id == id)
        if options:
            query = query.options(options)
        return query.one_or_none()

    def by_ids(self, ids, options=None) -> List[Task]:
        query = self.query(Task).filter(Task.id.in_(ids))
        if options:
            query = query.options(options)
        return query.all()

    def change_status(self, task, status: TaskStatus):
        if status == TaskStatus.InProgress:
            task.started = now()
        elif status in [
            TaskStatus.Failed, TaskStatus.Stopped, TaskStatus.Success
        ]:
            task.finished = now()

        task.status = status.value
        self.update()
        self.commit()

    def change_status_all(self, tasks: List[int], status: TaskStatus):
        updates = {'status': status.value}
        if status == TaskStatus.InProgress:
            updates['started'] = now()
        elif status in [
            TaskStatus.Failed, TaskStatus.Stopped, TaskStatus.Success
        ]:
            updates['finished'] = now()

        self.query(Task). \
            filter(Task.id.in_(tasks)). \
            update(updates,
                   synchronize_session=False)
        self.commit()

    def by_status(
            self,
            *statuses: TaskStatus,
            task_docker_assigned: str = None,
            worker_index: int = None,
            computer_assigned: str = None,
            project: int = None
    ):
        statuses = [s.value for s in statuses]
        query = self.query(Task).filter(Task.status.in_(statuses)). \
            options(joinedload(Task.dag_rel, innerjoin=True))

        if task_docker_assigned:
            query = query.filter(Task.docker_assigned == task_docker_assigned)
        if worker_index is not None:
            query = query.filter(Task.worker_index == worker_index)
        if computer_assigned is not None:
            query = query.filter(Task.computer_assigned == computer_assigned)
        if project:
            query = query.filter(Dag.project == project)
        return query.all()

    def dependency_status(self, tasks: List[Task]):
        res = {t.id: set() for t in tasks}
        task_ids = [task.id for task in tasks]
        items = self.query(TaskDependence, Task). \
            filter(TaskDependence.task_id.in_(task_ids)). \
            join(Task, Task.id == TaskDependence.depend_id).all()
        for item, task in items:
            res[item.task_id].add(task.status)

        return res

    def update_last_activity(self, task: int):
        self.query(Task).filter(Task.id == task
                                ).update({'last_activity': now()})
        self.session.commit()

    def stop(self, id: int = None, tasks: List[Task] = None):
        if id is not None:
            task = self.by_id(id)
            self.change_status(task, TaskStatus.Stopped)
        if tasks is not None:
            self.change_status_all(tasks=[t.id for t in tasks],
                                   status=TaskStatus.Stopped)

    def last_succeed_time(self):
        res = self.query(Task.finished). \
            filter(Task.status == TaskStatus.Success.value). \
            order_by(Task.finished.desc()). \
            first()
        return res[0] if res else None

    def by_dag(self, id: int):
        return self.query(Task).filter(Task.dag == id).order_by(Task.id).all()

    def parent_tasks_stats(self):
        task_parent = aliased(Task)
        task_child = aliased(Task)

        task_status = []
        for e in TaskStatus:
            task_status.append(
                func.sum(
                    case(whens=[(task_child.status == e.value, 1)],
                         else_=0).label(e.name)
                )
            )

        times = [func.min(task_child.started), func.max(task_child.finished)]

        parent_statuses = [
            TaskStatus.Queued.value, TaskStatus.InProgress.value
        ]

        query = self.query(task_parent, *times, *task_status). \
            filter(task_parent.status.in_(parent_statuses)). \
            filter(task_child.continued.__eq__(False)). \
            join(task_child, task_parent.id == task_child.parent,
                 isouter=True). \
            group_by(task_parent.id)

        res = []
        for task, started, finished, *(statuses) in query.all():
            res.append(
                [
                    task, started, finished,
                    {e: s
                     for e, s in zip(TaskStatus, statuses)}
                ]
            )

        return res

    def has_id(self, id: int):
        return self.query(Task).filter(Task.id == id).count() > 0

    def children(self, ids: Union[int, List[int]], joined_load=None):
        if isinstance(ids, int):
            ids = [ids]

        res = self.query(Task).filter(Task.parent.in_(ids)).filter(
            Task.continued.__eq__(False))

        res = res.order_by(Task.id)
        if joined_load is not None:
            for n in joined_load:
                res = res.options(joinedload(n, innerjoin=True))
        return res.all()

    def project(self, task_id: int):
        return self.query(Project).join(Dag).join(Task).filter(
            Task.id == task_id
        ).one()

    def find_dependents(self, task_id: int):
        res = self.query(Task). \
            join(TaskDependence, Task.id == TaskDependence.depend_id). \
            filter(TaskDependence.task_id == task_id). \
            all()
        return res


__all__ = ['TaskProvider']



from abc import ABC, abstractmethod
import time

from tqdm import tqdm

from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp import FILE_SYNC_INTERVAL
from mlcomp.db.core import Session
from mlcomp.db.models import Task, Dag
from mlcomp.utils.config import Config
from mlcomp.db.providers import TaskProvider, TaskSyncedProvider
from mlcomp.utils.misc import to_snake
from mlcomp.worker.executors.base.step import StepWrap


class TqdmWrapper:
    def __init__(
            self, executor: 'Executor', iterable=None,
            desc: str = 'progress', interval: int = 10,
            **kwargs
    ):
        self.desc = desc
        self.iterable = iterable
        self.interval = interval
        self.executor = executor
        self.tqdm = tqdm(iterable=iterable, **kwargs)

    def refresh(self):
        executor = self.executor
        tqdm = self.tqdm

        executor.task.loader_name = self.desc
        executor.task.batch_index = tqdm.n
        executor.task.batch_total = tqdm.total
        executor.task.epoch_duration = time.time() - tqdm.start_t
        if tqdm.n > 0:
            frac = (tqdm.total - tqdm.n) / tqdm.n
            executor.task.epoch_time_remaining = \
                executor.task.epoch_duration * frac

        executor.task_provider.update()
        return time.time()

    def set_description(self, desc=None, refresh=True):
        """
        Set/modify description of the progress bar.

        Parameters
        ----------
        desc  : str, optional
        refresh  : bool, optional
            Forces refresh [default: True].
        """
        self.desc = desc or ''
        if refresh:
            self.refresh()

    def __iter__(self):
        last_written = self.refresh()

        for item in self.tqdm:
            elapsed = time.time() - last_written
            if elapsed > self.interval:
                last_written = self.refresh()
            yield item
        self.refresh()


class Executor(ABC):
    _child = dict()

    session = None
    task_provider = None
    logger = None
    logger_db = None
    step = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def debug(self, message: str, db: bool = False):
        if self.step:
            self.step.debug(message, db=db)

    def info(self, message: str, db: bool = False):
        if self.step:
            self.step.info(message, db=db)

    def warning(self, message: str, db: bool = False):
        if self.step:
            self.step.warning(message, db=db)

    def error(self, message: str, db: bool = False):
        if self.step:
            self.step.error(message, db=db)

    def write(self, message: str):
        if message.strip() != '':
            self.info(message, db=True)

    def flush(self):
        pass

    def add_child_process(self, pid: int):
        additional_info = yaml_load(self.task.additional_info)
        additional_info['child_processes'] = additional_info.get(
            'child_processes', []) + [pid]
        self.task.additional_info = yaml_dump(additional_info)
        self.task_provider.update()

    def __call__(
            self, *, task: Task, task_provider: TaskProvider, dag: Dag
    ) -> dict:
        assert dag is not None, 'You must fetch task with dag_rel'

        self.task_provider = task_provider
        self.task = task
        self.dag = dag
        self.step = StepWrap(self.session, self.logger, self.logger_db, task,
                             task_provider)
        self.step.enter()

        if not task.debug and FILE_SYNC_INTERVAL:
            self.wait_data_sync()
        res = self.work()
        self.step.task_provider.commit()
        return res

    @abstractmethod
    def work(self) -> dict:
        pass

    @classmethod
    def _from_config(
            cls, executor: dict, config: Config, additional_info: dict
    ):
        return cls(**executor)

    @staticmethod
    def from_config(
            *, executor: str, config: Config, additional_info: dict,
            session: Session, logger, logger_db
    ) -> 'Executor':
        if executor not in config['executors']:
            raise ModuleNotFoundError(
                f'Executor {executor} '
                f'has not been found'
            )

        executor = additional_info['executor']
        child_class = Executor._child[executor['type']]

        # noinspection PyProtectedMember
        res = child_class._from_config(executor, config, additional_info)
        res.session = session
        res.logger = logger
        res.logger_db = logger_db
        return res

    @staticmethod
    def register(cls):
        Executor._child[cls.__name__] = cls
        Executor._child[cls.__name__.lower()] = cls
        Executor._child[to_snake(cls.__name__)] = cls
        return cls

    @staticmethod
    def is_registered(cls: str):
        return cls in Executor._child

    def wait_data_sync(self):
        self.info(f'Start data sync')

        while True:
            provider = TaskSyncedProvider(self.session)
            provider.commit()

            wait = False
            for computer, project, tasks in provider.for_computer(
                    self.task.computer_assigned
            ):
                if project.id == self.dag.project:
                    wait = True

            if wait:
                time.sleep(1)
            else:
                break

        self.info(f'Finish data sync')

    @staticmethod
    def is_trainable(type: str):
        variants = ['Catalyst']
        return type in (variants + [v.lower() for v in variants])

    def tqdm(self, iterable=None, desc: str = 'progress', interval: int = 10,
             **kwargs):
        """
        tqdm wrapper. writes progress to Database
        Args:
            iterable: iterable for tqdm
            desc: name of the progress bar
            interval: interval of writing to Database
            **kwargs: tqdm additional arguments

        Returns: tqdm wrapper
        """
        return TqdmWrapper(
            executor=self, iterable=iterable,
            desc=desc, interval=interval,
            **kwargs)

    def dependent_results(self):
        tasks = self.task_provider.find_dependents(self.task.id)
        res = dict()
        for t in tasks:
            res[t.id] = yaml_load(t.result)
        return res


__all__ = ['Executor']



import base64
import pickle
from itertools import groupby
from typing import List

from sqlalchemy import func, case
from sqlalchemy.orm import joinedload

from mlcomp.db.core import PaginatorOptions, Session
from mlcomp.db.enums import TaskStatus, TaskType
from mlcomp.db.models import Report, ReportTasks, Task, ReportSeries, \
    ReportImg, ReportLayout
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.db.report_info import ReportLayoutSeries, ReportLayoutInfo
from mlcomp.db.report_info.item import ReportLayoutItem
from mlcomp.utils.io import yaml_load, yaml_dump


class ReportProvider(BaseDataProvider):
    model = Report

    def __init__(self, session: Session = None):
        super(ReportProvider, self).__init__(session)

    def get(self, filter: dict, options: PaginatorOptions):
        task_count_cond = func.sum(
            case(
                whens=[(Task.status <= TaskStatus.InProgress.value, 1)],
                else_=0
            ).label('tasks_not_finished')
        )

        query = self.query(Report,
                           func.count(ReportTasks.task).label('tasks_count'),
                           task_count_cond, ). \
            join(ReportTasks, ReportTasks.report == Report.id, isouter=True). \
            join(Task, Task.id == ReportTasks.task, isouter=True)

        if filter.get('task'):
            query = query.filter(ReportTasks.task == filter['task'])

        query = query.group_by(Report.id)

        total = query.count()
        data = []
        for report, task_count, tasks_not_finished in self.paginator(
            query, options
        ):
            item = {
                'id': report.id,
                'time': self.serialize_datetime(report.time),
                'tasks': task_count,
                'tasks_not_finished': tasks_not_finished,
                'name': report.name
            }
            data.append(item)

        return {'total': total, 'data': data}

    def _detail_series(
        self, series: List[ReportSeries], r: ReportLayoutSeries
    ):
        series = [s for s in series if s.name == r.name]
        res = []

        series = sorted(series, key=lambda x: x.part)
        series_group = groupby(series, key=lambda x: x.part)
        for key, group in series_group:
            group = list(group)
            group = sorted(group, key=lambda x: x.task)
            for task_key, group_task in groupby(group, key=lambda x: x.task):
                group_task = list(group_task)
                res.append(
                    {
                        'x': [item.epoch for item in group_task],
                        'y': [item.value for item in group_task],
                        'stage': [item.stage for item in group_task],
                        'color': 'orange' if key == 'valid' else 'blue',
                        'time': [
                            self.serialize_datetime(item.time)
                            for item in group_task
                        ],
                        'group': key,
                        'task_name': group_task[0].task_rel.name,
                        'task_id': task_key,
                        'source': r.key,
                        'name': r.name
                    }
                )

        return res

    def _detail_single_img(self, report: int, item: ReportLayoutItem):
        res = []
        img_objs = self.query(ReportImg). \
            filter(ReportImg.group == item.name).all()

        for img_obj in img_objs:
            img_decoded = pickle.loads(img_obj.img)
            item = {
                'name': f'{img_obj.group} - {img_obj.part}',
                'data': base64.b64encode(img_decoded['img']).decode('utf-8')
            }
            res.append(item)

        return res

    def detail_img_classify_descr(self, report: int, item: ReportLayoutItem):
        res = []
        tasks = self.query(ReportTasks.task.distinct()
                           ).filter(ReportTasks.report == report).all()
        tasks = [t[0] for t in tasks]
        task_names = {
            id: name
            for id, name in self.query(Task.id, Task.name
                                       ).filter(Task.id.in_(tasks)).all()
        }
        for task in tasks:
            obj = {
                'name': task_names[task],
                'group': item.name,
                'task': task,
            }
            res.append(obj)
        return res

    def detail_img_segment_descr(self, report: int, item: ReportLayoutItem):
        res = []
        tasks = self.query(ReportTasks.task.distinct()
                           ).filter(ReportTasks.report == report).all()
        tasks = [t[0] for t in tasks]
        task_names = {
            id: name
            for id, name in self.query(Task.id, Task.name
                                       ).filter(Task.id.in_(tasks)).all()
        }
        for task in tasks:
            obj = {
                'name': task_names[task],
                'group': item.name,
                'task': task,
            }
            res.append(obj)
        return res

    def detail(self, id: int):
        report_obj = self.by_id(id)
        tasks = self.query(ReportTasks.task).filter(ReportTasks.report == id
                                                    ).all()
        tasks = [t[0] for t in tasks]
        config = yaml_load(report_obj.config)
        report = ReportLayoutInfo(config)

        series = self.query(ReportSeries). \
            filter(ReportSeries.task.in_(tasks)). \
            order_by(ReportSeries.epoch). \
            options(joinedload(ReportSeries.task_rel, innerjoin=True)).all()

        items = dict()
        for s in report.series:
            items[s.name] = self._detail_series(series, s)

        for element in report.precision_recall + report.f1:
            items[element.name] = self._detail_single_img(id, element)

        for element in report.img_classify:
            items[element.name] = self.detail_img_classify_descr(id, element)

        for element in report.img_segment:
            items[element.name] = self.detail_img_segment_descr(id, element)

        return {
            'data': items,
            'layout': report.layout,
            'metric': report.metric.serialize()
        }

    def add_dag(self, dag: int, report: int):
        tasks = self.query(Task.id). \
            filter(Task.dag == dag). \
            filter(Task.type <= TaskType.Train.value). \
            all()

        report_tasks = self.query(ReportTasks.task
                                  ).filter(ReportTasks.report == report).all()

        for t in set(t[0] for t in tasks) - set(t[0] for t in report_tasks):
            self.add(ReportTasks(report=report, task=t))

    def remove_dag(self, dag: int, report: int):
        tasks = self.query(Task.id).filter(Task.dag == dag).all()
        tasks = [t[0] for t in tasks]
        self.query(ReportTasks).filter(ReportTasks.report == report). \
            filter(ReportTasks.task.in_(tasks)). \
            delete(synchronize_session=False)

        self.session.commit()

    def remove_task(self, task: int, report: int):
        self.query(ReportTasks).filter(ReportTasks.report == report). \
            filter(ReportTasks.task == task). \
            delete(synchronize_session=False)

        self.session.commit()

    def add_task(self, task: int, report: int):
        self.add(ReportTasks(task=task, report=report))
        self.session.commit()

    def update_layout_start(self, id: int):
        layouts = self.query(ReportLayout.name).all()
        report = self.by_id(id)
        layouts = [l[0] for l in layouts]
        if report.layout in layouts:
            layouts.remove(report.layout)
            layouts.insert(0, report.layout)

        return {'id': id, 'layouts': layouts}

    def update_layout_end(self, id: int, layout: str, layouts: dict):
        layout_content = yaml_dump(layouts[layout])
        report = self.by_id(id)
        report.config = layout_content
        report.layout = layout
        self.commit()


__all__ = ['ReportProvider']



import torch.nn as nn

from ..common.blocks import Conv2dReLU
from ..base.model import Model


class TransposeX2(nn.Module):

    def __init__(self, in_channels, out_channels, use_batchnorm=True,
                 **batchnorm_params):
        super().__init__()
        layers = []
        layers.append(
            nn.ConvTranspose2d(in_channels, out_channels, kernel_size=4,
                               stride=2, padding=1))
        if use_batchnorm:
            layers.append(nn.BatchNorm2d(out_channels, **batchnorm_params))
        layers.append(nn.ReLU(inplace=True))

        self.block = nn.Sequential(*layers)

    def forward(self, x):
        return self.block(x)


class DecoderBlock(nn.Module):
    def __init__(self, in_channels, out_channels, use_batchnorm=True):
        super().__init__()

        self.block = nn.Sequential(
            Conv2dReLU(in_channels, in_channels // 4, kernel_size=1,
                       use_batchnorm=use_batchnorm),
            TransposeX2(in_channels // 4, in_channels // 4,
                        use_batchnorm=use_batchnorm),
            Conv2dReLU(in_channels // 4, out_channels, kernel_size=1,
                       use_batchnorm=use_batchnorm),
        )

    def forward(self, x):
        x, skip = x
        x = self.block(x)
        if skip is not None:
            x = x + skip
        return x


class LinknetDecoder(Model):

    def __init__(
            self,
            encoder_channels,
            prefinal_channels=32,
            final_channels=1,
            use_batchnorm=True,
    ):
        super().__init__()

        in_channels = encoder_channels

        self.layer1 = DecoderBlock(in_channels[0], in_channels[1],
                                   use_batchnorm=use_batchnorm)
        self.layer2 = DecoderBlock(in_channels[1], in_channels[2],
                                   use_batchnorm=use_batchnorm)
        self.layer3 = DecoderBlock(in_channels[2], in_channels[3],
                                   use_batchnorm=use_batchnorm)
        self.layer4 = DecoderBlock(in_channels[3], in_channels[4],
                                   use_batchnorm=use_batchnorm)
        self.layer5 = DecoderBlock(in_channels[4], prefinal_channels,
                                   use_batchnorm=use_batchnorm)
        self.final_conv = nn.Conv2d(prefinal_channels, final_channels,
                                    kernel_size=(1, 1))

        self.initialize()

    def forward(self, x):
        encoder_head = x[0]
        skips = x[1:]

        x = self.layer1([encoder_head, skips[0]])
        x = self.layer2([x, skips[1]])
        x = self.layer3([x, skips[2]])
        x = self.layer4([x, skips[3]])
        x = self.layer5([x, None])
        x = self.final_conv(x)

        return x



import pytest
from importlib import reload
import shutil

import mlcomp
from mlcomp import ROOT_FOLDER
from mlcomp.db.core import Session
from mlcomp.migration.manage import migrate


@pytest.fixture()
def session():
    if ROOT_FOLDER:
        shutil.rmtree(ROOT_FOLDER)
        reload(mlcomp)

    migrate()
    res = Session.create_session()
    yield res



from sqlalchemy.orm import joinedload

from mlcomp.db.core import Session
from mlcomp.db.models import Task, Model
from mlcomp.db.providers import TaskProvider, ProjectProvider, ModelProvider
from mlcomp.server.back.create_dags.standard import dag_standard
from mlcomp.utils.misc import now


def dag_model_add(session: Session, data: dict):
    if not data.get('task'):
        model = Model(
            name=data['name'],
            project=data['project'],
            equations=data['equations'],
            created=now()
        )
        ModelProvider(session).add(model)
        return

    task_provider = TaskProvider(session)
    task = task_provider.by_id(
        data['task'], options=joinedload(Task.dag_rel, innerjoin=True)
    )
    child_tasks = task_provider.children(task.id)
    computer = task.computer_assigned
    child_task = None
    if len(child_tasks) > 0:
        child_task = child_tasks[0].id
        computer = child_tasks[0].computer_assigned

    project = ProjectProvider(session).by_id(task.dag_rel.project)
    config = {
        'info': {
            'name': 'model_add',
            'project': project.name,
            'computer': computer
        },
        'executors': {
            'model_add': {
                'type': 'model_add',
                'project': data['project'],
                'task': data.get('task'),
                'name': data['name'],
                'file': data['file'],
                'child_task': child_task,
                'fold': data['fold']
            }
        }
    }

    dag_standard(
        session=session, config=config, debug=False, upload_files=False
    )


__all__ = ['dag_model_add']



from mlcomp.db.models import DagStorage, File, DagLibrary
from mlcomp.db.providers.base import BaseDataProvider


class DagStorageProvider(BaseDataProvider):
    model = DagStorage

    def by_dag(self, dag: int):
        query = self.query(DagStorage, File).join(File, isouter=True). \
            filter(DagStorage.dag == dag). \
            order_by(DagStorage.path)
        return query.all()


class DagLibraryProvider(BaseDataProvider):
    model = DagLibrary

    def dag(self, dag: int):
        return self.query(DagLibrary.library, DagLibrary.version). \
            filter(DagLibrary.dag == dag).all()


__all__ = ['DagStorageProvider', 'DagLibraryProvider']



import sqlalchemy as sa
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from mlcomp.db.models.base import Base
from mlcomp.utils.misc import now


class Dag(Base):
    __tablename__ = 'dag'

    id = sa.Column(sa.Integer, primary_key=True)
    project = sa.Column(sa.Integer, ForeignKey('project.id'))
    created = sa.Column(sa.DateTime, default=now())
    config = sa.Column(sa.String)
    name = sa.Column(sa.String)
    tasks = relationship('Task', lazy='noload')
    project_rel = relationship('Project', lazy='noload')
    docker_img = sa.Column(sa.String)
    img_size = sa.Column(sa.BigInteger, nullable=False, default=0)
    file_size = sa.Column(sa.BigInteger, nullable=False, default=0)
    type = sa.Column(sa.Integer, default=0)
    report = sa.Column(sa.Integer, ForeignKey('report.id'))
    report_rel = relationship('Report', lazy='noload')


__all__ = ['Dag']



import os
from multiprocessing import cpu_count

import click

from mlcomp import CONFIG_FOLDER, REDIS_PORT, REDIS_PASSWORD
from mlcomp.report import check_statuses
from mlcomp.server.back.app import start_server as _start_server
from mlcomp.server.back.app import stop_server as _stop_server
from mlcomp.utils.misc import kill_child_processes


@click.group()
def main():
    pass


@main.command()
def start_site():
    """
    Start only site
    """
    check_statuses()
    _start_server()


@main.command()
def stop_site():
    """
    Stop site
    """
    check_statuses()
    _stop_server()


@main.command()
@click.option('--daemon', type=bool, default=False,
              help='start supervisord in a daemon mode')
@click.option('--debug', type=bool, default=False,
              help='use source files instead the installed library')
@click.option('--workers', type=int, default=cpu_count(),
              help='count of workers')
@click.option('--log_level', type=str, default='DEBUG',
              help='log level of supervisord')
def start(daemon: bool, debug: bool, workers: int, log_level: str):
    """
    Start both server and worker on the same machine.

    It starts: redis-server, site, worker_supervisor, workers
    """
    check_statuses()

    # creating supervisord config
    supervisor_command = 'mlcomp-worker worker-supervisor'
    worker_command = 'mlcomp-worker worker'
    server_command = 'mlcomp-server start-site'

    if debug:
        supervisor_command = 'python mlcomp/worker/__main__.py ' \
                             'worker-supervisor'
        worker_command = 'python mlcomp/worker/__main__.py worker'
        server_command = 'python mlcomp/server/__main__.py start-site'

    folder = os.path.dirname(os.path.dirname(__file__))
    redis_path = os.path.join(folder, 'bin/redis-server')

    daemon_text = 'false' if daemon else 'true'
    text = [
        '[supervisord]', f'nodaemon={daemon_text}', '',
        '[program:supervisor]',
        f'command={supervisor_command}', 'autostart=true', 'autorestart=true',
        '', '[program:redis]', f'command={redis_path} --port {REDIS_PORT}'
                               f' --requirepass {REDIS_PASSWORD}',
        'autostart=true',
        'autorestart=true', '',
        '[program:server]',
        f'command={server_command}',
        'autostart=true',
        'autorestart=true', ''
    ]

    for p in range(workers):
        text.append(f'[program:worker{p}]')
        text.append(f'command={worker_command} {p}')
        text.append('autostart=true')
        text.append('autorestart=true')
        text.append('')

    conf = os.path.join(CONFIG_FOLDER, 'supervisord.conf')
    with open(conf, 'w') as f:
        f.writelines('\n'.join(text))

    os.system(f'supervisord ' f'-c {conf} -e {log_level}')


@main.command()
def stop():
    """
    Stop supervisord started by start command
    """
    check_statuses()

    lines = os.popen('ps -ef | grep supervisord').readlines()
    for line in lines:
        if 'mlcomp/configs/supervisord.conf' not in line:
            continue
        pid = int(line.split()[1])
        kill_child_processes(pid)


@main.command()
def status():
    """
    Shows if mlcomp server is already running
    """
    lines = os.popen("ps ef | grep mlcomp").readlines()
    pids = {}
    for line in lines:
        if "mlcomp/configs/supervisord.conf" in line:
            pids["server"] = line
        elif "mlcomp-server start-site" in line:
            pids["site"] = line
        elif "redis-server" in line:
            pids["redis"] = line
    if not pids:
        print("There are no mlcomp services started")
        return
    text = "Current MLComp services status:\n"
    for k, v in pids.items():
        text += f"  (✔) {k} is started on pid {v.split()[0]}\n"
    print(text)


if __name__ == '__main__':
    main()



from glob import glob
import os

from mlcomp.db.core import Session
from mlcomp.db.models import ReportLayout
from mlcomp.db.providers import ReportLayoutProvider
from mlcomp.utils.misc import now


def upgrade(migrate_engine):
    folder = os.path.dirname(__file__)
    session = Session.create_session(connection_string=migrate_engine.url)
    provider = ReportLayoutProvider(session)

    try:
        files = os.path.join(folder, '002', 'report_layout', '*.yml')
        for path in glob(files):
            name = str(os.path.basename(path).split('.')[0])
            text = open(path).read()
            provider.add(
                ReportLayout(name=name, content=text, last_modified=now()),
                commit=False
            )

        provider.commit()
    except Exception:
        provider.rollback()
        raise


def downgrade(migrate_engine):
    session = Session.create_session(connection_string=migrate_engine.url)
    provider = ReportLayoutProvider(session)
    provider.session.query(ReportLayout).delete(synchronize_session=False)
    provider.session.commit()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging

import torch.nn as nn


def r2plus1_unit(
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        groups,
        inplace_relu,
        bn_eps,
        bn_mmt,
        dim_mid=None,
):
    """
    Implementation of `R(2+1)D unit <https://arxiv.org/abs/1711.11248>`_.
    Decompose one 3D conv into one 2D spatial conv and one 1D temporal conv.
    Choose the middle dimensionality so that the total No. of parameters
    in 2D spatial conv and 1D temporal conv is unchanged.

    Args:
        dim_in (int): the channel dimensions of the input.
        dim_out (int): the channel dimension of the output.
        temporal_stride (int): the temporal stride of the bottleneck.
        spatial_stride (int): the spatial_stride of the bottleneck.
        groups (int): number of groups for the convolution.
        inplace_relu (bool): calculate the relu on the original input
            without allocating new memory.
        bn_eps (float): epsilon for batch norm.
        bn_mmt (float): momentum for batch norm. Noted that BN momentum in
            PyTorch = 1 - BN momentum in Caffe2.
        dim_mid (Optional[int]): If not None, use the provided channel dimension
            for the output of the 2D spatial conv. If None, compute the output
            channel dimension of the 2D spatial conv so that the total No. of
            model parameters remains unchanged.
    """
    if dim_mid is None:
        dim_mid = int(
            dim_out * dim_in * 3 * 3 * 3 / (dim_in * 3 * 3 + dim_out * 3))
        logging.info(
            "dim_in: %d, dim_out: %d. Set dim_mid to %d" % (
            dim_in, dim_out, dim_mid)
        )
    # 1x3x3 group conv, BN, ReLU
    conv_middle = nn.Conv3d(
        dim_in,
        dim_mid,
        [1, 3, 3],  # kernel
        stride=[1, spatial_stride, spatial_stride],
        padding=[0, 1, 1],
        groups=groups,
        bias=False,
    )
    conv_middle_bn = nn.BatchNorm3d(dim_mid, eps=bn_eps, momentum=bn_mmt)
    conv_middle_relu = nn.ReLU(inplace=inplace_relu)
    # 3x1x1 group conv
    conv = nn.Conv3d(
        dim_mid,
        dim_out,
        [3, 1, 1],  # kernel
        stride=[temporal_stride, 1, 1],
        padding=[1, 0, 0],
        groups=groups,
        bias=False,
    )
    return nn.Sequential(conv_middle, conv_middle_bn, conv_middle_relu, conv)



import numpy as np

import torch
import torch.nn.functional as F

from pretrainedmodels.models.dpn import DPN
from pretrainedmodels.models.dpn import pretrained_settings


class DPNEncorder(DPN):

    def __init__(self, feature_blocks, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.feature_blocks = np.cumsum(feature_blocks)
        self.pretrained = False

        del self.last_linear

    def forward(self, x):

        features = []

        input_block = self.features[0]

        x = input_block.conv(x)
        x = input_block.bn(x)
        x = input_block.act(x)
        features.append(x)

        x = input_block.pool(x)

        for i, module in enumerate(self.features[1:], 1):
            x = module(x)
            if i in self.feature_blocks:
                features.append(x)

        out_features = [
            features[4],
            F.relu(torch.cat(features[3], dim=1), inplace=True),
            F.relu(torch.cat(features[2], dim=1), inplace=True),
            F.relu(torch.cat(features[1], dim=1), inplace=True),
            features[0],
        ]

        return out_features

    def load_state_dict(self, state_dict, **kwargs):
        state_dict.pop('last_linear.bias')
        state_dict.pop('last_linear.weight')
        super().load_state_dict(state_dict, **kwargs)


dpn_encoders = {
    'dpn68': {
        'encoder': DPNEncorder,
        'out_shapes': (832, 704, 320, 144, 10),
        'pretrained_settings': pretrained_settings['dpn68'],
        'params': {
            'feature_blocks': (3, 4, 12, 4),
            'groups': 32,
            'inc_sec': (16, 32, 32, 64),
            'k_r': 128,
            'k_sec': (3, 4, 12, 3),
            'num_classes': 1000,
            'num_init_features': 10,
            'small': True,
            'test_time_pool': True
        },
    },

    'dpn68b': {
        'encoder': DPNEncorder,
        'out_shapes': (832, 704, 320, 144, 10),
        'pretrained_settings': pretrained_settings['dpn68b'],
        'params': {
            'feature_blocks': (3, 4, 12, 4),
            'b': True,
            'groups': 32,
            'inc_sec': (16, 32, 32, 64),
            'k_r': 128,
            'k_sec': (3, 4, 12, 3),
            'num_classes': 1000,
            'num_init_features': 10,
            'small': True,
            'test_time_pool': True,
        },
    },

    'dpn92': {
        'encoder': DPNEncorder,
        'out_shapes': (2688, 1552, 704, 336, 64),
        'pretrained_settings': pretrained_settings['dpn92'],
        'params': {
            'feature_blocks': (3, 4, 20, 4),
            'groups': 32,
            'inc_sec': (16, 32, 24, 128),
            'k_r': 96,
            'k_sec': (3, 4, 20, 3),
            'num_classes': 1000,
            'num_init_features': 64,
            'test_time_pool': True
        },
    },

    'dpn98': {
        'encoder': DPNEncorder,
        'out_shapes': (2688, 1728, 768, 336, 96),
        'pretrained_settings': pretrained_settings['dpn98'],
        'params': {
            'feature_blocks': (3, 6, 20, 4),
            'groups': 40,
            'inc_sec': (16, 32, 32, 128),
            'k_r': 160,
            'k_sec': (3, 6, 20, 3),
            'num_classes': 1000,
            'num_init_features': 96,
            'test_time_pool': True,
        },
    },

    'dpn107': {
        'encoder': DPNEncorder,
        'out_shapes': (2688, 2432, 1152, 376, 128),
        'pretrained_settings': pretrained_settings['dpn107'],
        'params': {
            'feature_blocks': (4, 8, 20, 4),
            'groups': 50,
            'inc_sec': (20, 64, 64, 128),
            'k_r': 200,
            'k_sec': (4, 8, 20, 3),
            'num_classes': 1000,
            'num_init_features': 128,
            'test_time_pool': True
        },
    },

    'dpn131': {
        'encoder': DPNEncorder,
        'out_shapes': (2688, 1984, 832, 352, 128),
        'pretrained_settings': pretrained_settings['dpn131'],
        'params': {
            'feature_blocks': (4, 8, 28, 4),
            'groups': 40,
            'inc_sec': (16, 32, 32, 128),
            'k_r': 160,
            'k_sec': (4, 8, 28, 3),
            'num_classes': 1000,
            'num_init_features': 128,
            'test_time_pool': True
        },
    },

}



from mlcomp.db.core import Session
from mlcomp.db.models import Dag
from mlcomp.db.providers import ModelProvider, DagProvider
from mlcomp.server.back.create_dags.standard import dag_standard
from mlcomp.utils.config import Config
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.utils.misc import now


def dag_model_start(session: Session, data: dict):
    provider = ModelProvider(session)
    model = provider.by_id(data['model_id'])
    dag_provider = DagProvider(session)
    dag = dag_provider.by_id(data['dag'], joined_load=[Dag.project_rel])

    project = dag.project_rel
    src_config = Config.from_yaml(dag.config)
    pipe = src_config['pipes'][data['pipe']['name']]

    equations = yaml_load(model.equations)
    versions = data['pipe']['versions']

    if len(versions) > 0:
        version = data['pipe']['version']
        pipe_equations = yaml_load(version['equations'])
        found_version = versions[0]
        for v in versions:
            if v['name'] == version['name']:
                found_version = v
                break

        found_version['used'] = now()

        for v in pipe.values():
            v.update(pipe_equations)

    equations[data['pipe']['name']] = versions
    model.equations = yaml_dump(equations)

    for v in pipe.values():
        v['model_id'] = model.id
        v['model_name'] = model.name

    config = {
        'info': {
            'name': data['pipe']['name'],
            'project': project.name
        },
        'executors': pipe
    }

    if model.dag:
        old_dag = dag_provider.by_id(model.dag)
        if old_dag.name != dag.name:
            model.dag = dag.id
    else:
        model.dag = dag.id

    provider.commit()

    dag_standard(
        session=session,
        config=config,
        debug=False,
        upload_files=False,
        copy_files_from=data['dag']
    )


__all__ = ['dag_model_start']



import base64
import pickle

import cv2
import numpy as np
from sqlalchemy import and_, or_

from mlcomp.db.core import PaginatorOptions
from mlcomp.db.models import Project, Dag, ReportImg, Task
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.img import resize_saving_ratio
from mlcomp.utils.io import yaml_load


class ReportImgProvider(BaseDataProvider):
    model = ReportImg

    def remove(self, filter: dict):
        query = self.query(ReportImg)
        if filter.get('dag'):
            query = query.filter(ReportImg.dag == filter['dag'])
        if filter.get('project'):
            query = query.filter(ReportImg.project == filter['project'])
        query.delete(synchronize_session=False)
        self.session.commit()

        query = self.query(Dag)
        if filter.get('dag'):
            query.filter(Dag.id == filter['dag']).update({'img_size': 0})

        if filter.get('project'):
            query.filter(Dag.project == filter['project']
                         ).update({'img_size': 0})

        self.session.commit()

    def remove_lower(self, task_id: int, name: str, epoch: int):
        self.query(ReportImg).filter(ReportImg.task == task_id). \
            filter(ReportImg.group == name). \
            filter(ReportImg.epoch < epoch). \
            delete(synchronize_session=False)
        self.session.commit()

    def detail_img_classify(
        self, filter: dict, options: PaginatorOptions = None
    ):
        res = {'data': []}
        confusion = self.query(ReportImg.img). \
            filter(ReportImg.task == filter['task']). \
            filter(ReportImg.group == filter['group'] + '_confusion').first()

        if confusion:
            confusion = pickle.loads(confusion[0])['data']
            res['confusion'] = {'data': confusion.tolist()}

        res.update(filter)

        query = self.query(ReportImg).filter(
            ReportImg.task == filter['task']). \
            filter(ReportImg.group == filter['group'])

        if filter.get('y') is not None and filter.get('y_pred') is not None:
            query = query.filter(
                and_(
                    ReportImg.y == filter['y'],
                    ReportImg.y_pred == filter['y_pred']
                )
            )

        if filter.get('score_min') is not None:
            query = query.filter(
                or_(
                    ReportImg.score >= filter['score_min'],
                    ReportImg.score.__eq__(None)
                )
            )

        if filter.get('score_max') is not None:
            query = query.filter(
                or_(
                    ReportImg.score <= filter['score_max'],
                    ReportImg.score.__eq__(None)
                )
            )

        layout = filter.get('layout')

        if layout and layout.get('attrs'):
            for attr in layout['attrs']:
                field = getattr(ReportImg, attr['source'])
                if attr.get('equal') is not None:
                    query = query.filter(field == attr['equal'])
                if attr.get('greater') is not None:
                    query = query.filter(field >= attr['greater'])
                if attr.get('less') is not None:
                    query = query.filter(field <= attr['less'])

        res['total'] = query.count()

        if confusion is not None:
            project = self.query(Project).join(Dag).join(Task).filter(
                Task.id == filter['task']
            ).first()
            class_names = yaml_load(project.class_names)

            if 'default' in class_names:
                res['class_names'] = class_names['default']
            else:
                res['class_names'] = [
                    str(i) for i in range(confusion.shape[1])
                ]

        query = self.paginator(query, options)
        img_objs = query.all()
        for img_obj in img_objs:
            buffer = img_obj.img
            if layout:
                buffer = np.fromstring(buffer, np.uint8)
                img = cv2.imdecode(buffer, cv2.IMREAD_COLOR)
                img = resize_saving_ratio(
                    img, (layout.get('max_height'), layout.get('max_width'))
                )
                retval, buffer = cv2.imencode('.jpg', img)

            jpg_as_text = base64.b64encode(buffer).decode('utf-8')

            # noinspection PyTypeChecker
            res['data'].append(
                {
                    'content': jpg_as_text,
                    'id': img_obj.id,
                    'y_pred': img_obj.y_pred,
                    'y': img_obj.y,
                    'score': round(img_obj.score, 2) if img_obj.score else None
                }
            )

        return res

    def detail_img_segment(
        self, filter: dict, options: PaginatorOptions = None
    ):
        res = {'data': []}
        res.update(filter)

        query = self.query(ReportImg).filter(
            ReportImg.task == filter['task']). \
            filter(ReportImg.group == filter['group'])

        if filter.get('y') is not None and filter.get('y_pred') is not None:
            query = query.filter(
                and_(
                    ReportImg.y == filter['y'],
                    ReportImg.y_pred == filter['y_pred']
                )
            )

        if filter.get('score_min') is not None:
            query = query.filter(
                or_(
                    ReportImg.score >= filter['score_min'],
                    ReportImg.score.__eq__(None)
                )
            )

        if filter.get('score_max') is not None:
            query = query.filter(
                or_(
                    ReportImg.score <= filter['score_max'],
                    ReportImg.score.__eq__(None)
                )
            )

        layout = filter.get('layout')

        if layout and layout.get('attrs'):
            for attr in layout['attrs']:
                field = getattr(ReportImg, attr['source'])
                if attr.get('equal') is not None:
                    query = query.filter(field == attr['equal'])
                if attr.get('greater') is not None:
                    query = query.filter(field >= attr['greater'])
                if attr.get('less') is not None:
                    query = query.filter(field <= attr['less'])

        res['total'] = query.count()

        query = self.paginator(query, options)
        img_objs = query.all()
        for img_obj in img_objs:
            buffer = img_obj.img
            if layout:
                buffer = np.fromstring(buffer, np.uint8)
                img = cv2.imdecode(buffer, cv2.IMREAD_COLOR)
                img = resize_saving_ratio(
                    img, (layout.get('max_height'), layout.get('max_width'))
                )
                retval, buffer = cv2.imencode('.jpg', img)

            jpg_as_text = base64.b64encode(buffer).decode('utf-8')

            # noinspection PyTypeChecker
            res['data'].append(
                {
                    'content': jpg_as_text,
                    'id': img_obj.id,
                    'y_pred': img_obj.y_pred,
                    'y': img_obj.y,
                    'score': round(img_obj.score, 2)
                    if img_obj.score is not None else None
                }
            )

        return res


__all__ = ['ReportImgProvider']



import torch.nn as nn


class Model(nn.Module):

    def __init__(self):
        super().__init__()

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out',
                                        nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)



from .project import Project
from .task import Task, TaskDependence, TaskSynced
from .file import File
from .dag_storage import DagStorage, DagLibrary
from .computer import Computer, ComputerUsage
from .log import Log
from .step import Step
from .dag import Dag
from .report import ReportSeries, ReportImg, ReportTasks, Report, ReportLayout
from .docker import Docker
from .model import Model
from .auxilary import Auxiliary

__all__ = [
    'Project', 'Task', 'TaskDependence', 'File', 'DagStorage', 'DagLibrary',
    'Computer', 'ComputerUsage', 'Log', 'Step', 'Dag', 'ReportSeries',
    'ReportImg', 'ReportTasks', 'Report', 'ReportLayout', 'Docker', 'Model',
    'Auxiliary', 'TaskSynced'
]



from torchvision.models.resnet import ResNet
from torchvision.models.resnet import BasicBlock
from torchvision.models.resnet import Bottleneck
from pretrainedmodels.models.torchvision_models import pretrained_settings


class ResNetEncoder(ResNet):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained = False
        del self.fc

    def forward(self, x):
        x0 = self.conv1(x)
        x0 = self.bn1(x0)
        x0 = self.relu(x0)

        x1 = self.maxpool(x0)
        x1 = self.layer1(x1)

        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        return [x4, x3, x2, x1, x0]

    def load_state_dict(self, state_dict, **kwargs):
        state_dict.pop('fc.bias')
        state_dict.pop('fc.weight')
        super().load_state_dict(state_dict, **kwargs)


resnet_encoders = {
    'resnet18': {
        'encoder': ResNetEncoder,
        'pretrained_settings': pretrained_settings['resnet18'],
        'out_shapes': (512, 256, 128, 64, 64),
        'params': {
            'block': BasicBlock,
            'layers': [2, 2, 2, 2],
        },
    },

    'resnet34': {
        'encoder': ResNetEncoder,
        'pretrained_settings': pretrained_settings['resnet34'],
        'out_shapes': (512, 256, 128, 64, 64),
        'params': {
            'block': BasicBlock,
            'layers': [3, 4, 6, 3],
        },
    },

    'resnet50': {
        'encoder': ResNetEncoder,
        'pretrained_settings': pretrained_settings['resnet50'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 6, 3],
        },
    },

    'resnet101': {
        'encoder': ResNetEncoder,
        'pretrained_settings': pretrained_settings['resnet101'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 23, 3],
        },
    },

    'resnet152': {
        'encoder': ResNetEncoder,
        'pretrained_settings': pretrained_settings['resnet152'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 8, 36, 3],
        },
    },

    'resnext50_32x4d': {
        'encoder': ResNetEncoder,
        'pretrained_settings': {
            'imagenet': {
                'url': 'https://download.pytorch.org/'
                       'models/resnext50_32x4d-7cdf4587.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            }
        },
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 6, 3],
            'groups': 32,
            'width_per_group': 4
        },
    },

    'resnext101_32x8d': {
        'encoder': ResNetEncoder,
        'pretrained_settings': {
            'imagenet': {
                'url': 'https://download.pytorch.org/'
                       'models/resnext101_32x8d-8ba56ff5.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            },
            'instagram': {
                'url': 'https://download.pytorch.org/'
                       'models/ig_resnext101_32x8-c38310e5.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            }
        },
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 23, 3],
            'groups': 32,
            'width_per_group': 8
        },
    },

    'resnext101_32x16d': {
        'encoder': ResNetEncoder,
        'pretrained_settings': {
            'instagram': {
                'url': 'https://download.pytorch.org/'
                       'models/ig_resnext101_32x16-c6f796b0.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            }
        },
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 23, 3],
            'groups': 32,
            'width_per_group': 16
        },
    },

    'resnext101_32x32d': {
        'encoder': ResNetEncoder,
        'pretrained_settings': {
            'instagram': {
                'url': 'https://download.pytorch.org/'
                       'models/ig_resnext101_32x32-e4b90b00.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            }
        },
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 23, 3],
            'groups': 32,
            'width_per_group': 32
        },
    },

    'resnext101_32x48d': {
        'encoder': ResNetEncoder,
        'pretrained_settings': {
            'instagram': {
                'url': 'https://download.pytorch.org/'
                       'models/ig_resnext101_32x48-3e41cc8a.pth',
                'input_space': 'RGB',
                'input_size': [3, 224, 224],
                'input_range': [0, 1],
                'mean': [0.485, 0.456, 0.406],
                'std': [0.229, 0.224, 0.225],
                'num_classes': 1000
            }
        },
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': Bottleneck,
            'layers': [3, 4, 23, 3],
            'groups': 32,
            'width_per_group': 48
        },
    },
}



from .project import ProjectProvider
from .task import TaskProvider
from .file import FileProvider
from .dag_storage import DagStorageProvider, DagLibraryProvider
from .log import LogProvider
from .step import StepProvider
from .computer import ComputerProvider
from .dag import DagProvider
from .report import \
    ReportImgProvider, \
    ReportProvider, \
    ReportLayoutProvider, \
    ReportSeriesProvider, \
    ReportTasksProvider
from .docker import DockerProvider
from .model import ModelProvider
from .auxiliary import AuxiliaryProvider
from .task_synced import TaskSyncedProvider

__all__ = [
    'ProjectProvider', 'TaskProvider', 'FileProvider', 'DagStorageProvider',
    'DagLibraryProvider', 'LogProvider', 'StepProvider', 'ComputerProvider',
    'DagProvider', 'ReportImgProvider', 'ReportProvider',
    'ReportLayoutProvider', 'ReportSeriesProvider', 'ReportTasksProvider',
    'DockerProvider', 'ModelProvider', 'AuxiliaryProvider',
    'TaskSyncedProvider'
]



import shutil
import traceback
from glob import glob
from os import makedirs
from os.path import dirname, join, basename, exists

import pandas as pd
import migrate.versioning.api as api
from mlcomp.utils.io import zip_folder

from mlcomp.db.enums import LogStatus, ComponentType

from mlcomp import SA_CONNECTION_STRING, REPORT_FOLDER, LOG_FOLDER, DB_TYPE, \
    CONFIG_FOLDER, ROOT_FOLDER, DATA_FOLDER, MODEL_FOLDER, TASK_FOLDER, \
    DB_FOLDER, TMP_FOLDER
from mlcomp.utils.misc import now, to_snake

from mlcomp.db.providers import DagProvider, LogProvider


def statuses(folder: str = None):
    rows = []

    folder_status = 'OK'
    folder_comment = ''

    folders = [
        ROOT_FOLDER,
        DATA_FOLDER,
        MODEL_FOLDER,
        TASK_FOLDER,
        LOG_FOLDER,
        CONFIG_FOLDER,
        DB_FOLDER,
        REPORT_FOLDER,
        TMP_FOLDER
    ]
    for f in folders:
        if not exists(f):
            folder_status = 'ERROR'
            folder_comment = f'folder {f} does not exist'

    files = [
        join(CONFIG_FOLDER, '.env')
    ]
    for f in files:
        if not exists(f):
            folder_status = 'ERROR'
            folder_comment = f'file {f} does not exist'

    rows.append({
        'name': 'Folders',
        'status': folder_status,
        'comment': folder_comment
    })

    database_status = 'OK'
    database_comment = f'DB_TYPE = {DB_TYPE}'
    try:
        provider = DagProvider()
        provider.count()
    except Exception:
        database_status = 'ERROR'
        database_comment += ' ' + traceback.format_exc()

    rows.append({
        'name': 'Database',
        'status': database_status,
        'comment': database_comment
    })

    if database_status == 'OK':
        migrate_status = 'OK'

        repository_folder = join(dirname(__file__), 'migration')
        repository_version = api.version(repository_folder)

        db_version = api.db_version(SA_CONNECTION_STRING, repository_folder)

        if db_version != repository_version:
            migrate_status = 'ERROR'
            migrate_comment = f'Repository version = {repository_version} ' \
                              f'Db version = {db_version}'
        else:
            migrate_comment = f'version: {db_version}'

        rows.append({
            'name': 'Migrate',
            'status': migrate_status,
            'comment': migrate_comment
        })

    df = pd.DataFrame(rows)

    if folder is not None:
        print('Statuses:')
        print(df)

        df.to_csv(join(folder, 'statuses.csv'), index=False)

    return df


def check_statuses():
    stats = statuses()
    failed = stats[stats['status'] != 'OK']
    if failed.shape[0] > 0:
        print('There are errors in statuses')
        for row in stats.itertuples():
            print(f'name: {row.name} status'
                  f' {row.status} comment {row.comment}')

        import time
        time.sleep(0.01)

        raise Exception('There are errors in statuses. '
                        'Please check them above')


def logs(folder: str = None):
    log_provider = LogProvider()
    errors = log_provider.last(count=1000, levels=[LogStatus.Error.value])
    service_components = [ComponentType.Supervisor.value,
                          ComponentType.API.value,
                          ComponentType.WorkerSupervisor.value]
    services = log_provider.last(count=1000, components=service_components)
    logs = errors + services

    rows = []
    for l, _ in logs:
        rows.append({
            'status': to_snake(LogStatus(l.level).name),
            'component': to_snake(ComponentType(l.component).name),
            'time': l.time,
            'message': l.message,
        })
    df = pd.DataFrame(rows)
    df.to_csv(join(folder, 'logs_db.csv'), index=False)

    if folder is not None:
        for file in glob(join(LOG_FOLDER, '*')):
            shutil.copy(file, join(folder, basename(file)))
        print('logs formed')
    return df


def create_report():
    print('*** Report Start ***')
    print()

    folder = join(REPORT_FOLDER, f'{now()}'.split('.')[0])
    makedirs(folder, exist_ok=True)

    statuses(folder)

    print()

    logs(folder)

    print()

    zip_path = folder + '.zip'
    zip_folder(folder, dst=zip_path)

    print('Report path', zip_path)
    print()

    print('*** Report End ***')


__all__ = ['check_statuses', 'create_report']



import json
import datetime
from collections import defaultdict

from sqlalchemy import func, case
import numpy as np

from mlcomp.db.core import PaginatorOptions
from mlcomp.db.enums import TaskStatus
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.db.models import Computer, ComputerUsage, Task, Docker, Project
from mlcomp.utils.misc import now, parse_time


class ComputerProvider(BaseDataProvider):
    model = Computer

    def computers(self):
        return {
            c.name:
                {k: v
                 for k, v in c.__dict__.items() if not k.startswith('_')}
            for c in self.query(Computer).all()
        }

    def get(self, filter: dict, options: PaginatorOptions = None):
        query = self.query(Computer)
        query = query.filter(
            Computer.last_synced >= now() - datetime.timedelta(days=1))
        total = query.count()
        if options:
            query = self.paginator(query, options)
        res = []
        for c in query.all():
            item = self.to_dict(c)
            default_usage = {
                'cpu': 0,
                'memory': 0,
                'gpu': [{
                    'memory': 0,
                    'load': 0
                } for i in range(item['gpu'])]
            }
            sync_status = 'Not synced'
            sync_date = None
            if c.last_synced:
                sync_date = self.serialize_datetime(c.last_synced)
                sync_status = f'Last synced'

            if c.syncing_computer:
                if c.last_synced is None or \
                        (now() - c.last_synced).total_seconds() >= 5:
                    sync_status = f'Syncing with {c.syncing_computer}'
                    if c.last_synced:
                        sync_status += f' from '
                    sync_date = self.serialize_datetime(c.last_synced)

            item['sync_status'] = sync_status
            item['sync_date'] = sync_date

            item['usage'] = json.loads(item['usage']) \
                if item['usage'] else default_usage
            item['memory'] = int(item['memory'] / 1000)
            item['usage']['cpu'] = int(item['usage']['cpu'])
            item['usage']['memory'] = int(item['usage']['memory'])
            for gpu in item['usage']['gpu']:
                gpu['memory'] = 0 if np.isnan(gpu['memory']) else int(
                    gpu['memory'])
                gpu['load'] = 0 if np.isnan(gpu['load']) else int(gpu['load'])

            min_time = parse_time(filter.get('usage_min_time'))

            item['usage_history'] = self.usage_history(
                c.name, min_time
            )
            item['dockers'] = self.dockers(c.name, c.cpu)
            res.append(item)

        return {'data': res, 'total': total}

    def usage_history(self, computer: str, min_time: datetime = None):
        min_time = min_time or (now() - datetime.timedelta(days=1))
        query = self.query(ComputerUsage).filter(
            ComputerUsage.time >= min_time
        ).filter(ComputerUsage.computer == computer
                 ).order_by(ComputerUsage.time)
        res = {'time': [], 'mean': []}
        mean = defaultdict(list)
        for c in query.all():
            item = self.to_dict(c, datetime_format='%Y-%m-%d %H:%M:%SZ')
            usage = json.loads(item['usage'])
            res['time'].append(item['time'])

            mean['cpu'].append(usage['mean']['cpu'])
            mean['memory'].append(usage['mean']['memory'])
            mean['disk'].append(usage['mean']['disk'])
            for i, gpu in enumerate(usage['mean']['gpu']):
                mean[f'gpu_{i}'].append(gpu['load'])

        for item in mean:
            res['mean'].append({'name': item, 'value': mean[item]})

        return dict(res)

    def current_usage(self, name: str, usage: dict):
        computer = self.query(Computer).filter(Computer.name == name).first()
        computer.usage = json.dumps(usage)
        self.update()

    def by_name(self, name: str):
        return self.query(Computer).filter(Computer.name == name).one()

    def computers_have_succeeded_tasks(self, min_time: datetime):
        res = self.session.query(Task.computer_assigned.distinct()). \
            filter(Task.finished >= min_time). \
            filter(Task.status == TaskStatus.Success.value). \
            all()
        res = [r[0] for r in res]
        return self.session.query(Computer). \
            filter(Computer.name.in_(res)). \
            all()

    def dockers(self, computer: str, cpu: int):
        count_cond = func.sum(
            case(
                whens=[(Task.status == TaskStatus.InProgress.value, 1)],
                else_=0
            ).label('count')
        )

        res = self.query(Docker, count_cond). \
            join(Task, Task.computer_assigned == computer, isouter=True). \
            filter(Docker.computer == computer). \
            group_by(Docker.name, Docker.computer). \
            all()
        return [
            {
                'name': r[0].name,
                'last_activity': self.serialize_datetime(
                    r[0].last_activity
                ),
                'in_progress': r[1],
                'free': cpu - r[1]
            } for r in res
        ]

    def all_with_last_activtiy(self):
        query = self.query(Computer, func.max(Docker.last_activity)). \
            join(Docker, Docker.computer == Computer.name). \
            group_by(Computer.name)
        res = []
        for c, a in query.all():
            c.last_activity = a
            res.append(c)
        return res

    def gpu_available(self):
        computers = self.all_with_last_activtiy()
        return sum([c.gpu for c in computers if
                    (now() - c.last_activity).total_seconds() <= 30])

    def sync_start(self):
        projects = self.query(Project).order_by(Project.id.desc()).all()
        res = []
        for p in projects:
            res.append({
                'id': p.id,
                'name': p.name,
                'sync_folders': p.sync_folders,
                'ignore_folders': p.ignore_folders,
            })
        return {'projects': res}


__all__ = ['ComputerProvider']



import sqlalchemy as sa
from sqlalchemy import ForeignKey

from mlcomp.db.models.base import Base


class Docker(Base):
    __tablename__ = 'docker'

    name = sa.Column(sa.String, primary_key=True)
    computer = sa.Column(sa.String,
                         ForeignKey('computer.name'),
                         primary_key=True)
    last_activity = sa.Column(sa.DateTime, nullable=False)
    ports = sa.Column(sa.String, nullable=False)


__all__ = ['Docker']



from efficientnet_pytorch import EfficientNet as _EfficientNet

import torch.nn as nn

from mlcomp.contrib.torch.layers import LambdaLayer


class EfficientNet(nn.Module):
    def __init__(self, variant, num_classes, pretrained=True, activation=None):
        super().__init__()
        if 'efficientnet' not in variant:
            variant = f'efficientnet-{variant}'

        if pretrained:
            model = _EfficientNet.from_pretrained(variant,
                                                  num_classes=num_classes)
        else:
            model = _EfficientNet.from_name(variant, {
                'num_classes': num_classes
            })
        self.model = model

        self.model._fc = nn.Sequential(
            LambdaLayer(lambda x: x.unsqueeze_(0)),
            nn.AdaptiveAvgPool1d(self.model._fc.in_features),
            LambdaLayer(lambda x: x.squeeze_(0).view(x.size(0), -1)),
            self.model._fc
        )

        if callable(activation) or activation is None:
            self.activation = activation
        elif activation == 'softmax':
            self.activation = nn.Softmax(dim=1)
        elif activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        else:
            raise ValueError(
                'Activation should be "sigmoid"/"softmax"/callable/None')

    def forward(self, x):
        res = self.model(x)
        if isinstance(res, tuple):
            res = res[0]
        if self.activation:
            res = self.activation(res)
        return res


__all__ = ['EfficientNet']



import time
from math import ceil
from socket import gethostname
import warnings
import datetime
from typing import List

from IPython import display
import matplotlib.pyplot as plt
import networkx as nx
from matplotlib.ticker import MaxNLocator
import pandas as pd

from mlcomp.db.enums import TaskStatus, ComponentType
from mlcomp.db.providers import TaskProvider, LogProvider, \
    DagProvider, ComputerProvider, ReportSeriesProvider
from mlcomp.utils.misc import now, to_snake

warnings.simplefilter('ignore')


def describe_tasks(dag: int, axis):
    provider = TaskProvider()
    columns = ['Id', 'Started', 'Duration', 'Step', 'Status']
    cells = []
    cells_colours = []

    tasks = provider.by_dag(dag)

    status_colors = {
        'not_ran': 'gray',
        'queued': 'lightblue',
        'in_progress': 'lime',
        'failed': '#e83217',
        'stopped': '#cb88ea',
        'skipped': 'orange',
        'success': 'green'
    }

    finish = True

    for task in tasks:
        started = ''
        duration = ''

        if task.status <= TaskStatus.InProgress.value:
            finish = False

        if task.started:
            started = task.started.strftime('%m.%d %H:%M:%S')
            if task.finished:
                duration = (task.finished - task.started).total_seconds()
            else:
                duration = (now() - task.started).total_seconds()

            if duration > 3600:
                duration = f'{int(duration // 3600)} hours ' \
                           f'{int((duration % 3600) // 60)} min' \
                           f' {int(duration % 60)} sec'
            elif duration > 60:
                duration = f'{int(duration // 60)} min' \
                           f' {int(duration % 60)} sec'
            else:
                duration = f'{int(duration)} sec'

        status = to_snake(TaskStatus(task.status).name)
        status_color = status_colors[status]

        task_cells = [
            str(task.id), started, duration, task.current_step or '1', status
        ]
        task_colors = ['white', 'white', 'white', 'white', status_color]
        cells.append(task_cells)
        cells_colours.append(task_colors)

    table = axis.table(
        cellText=cells,
        colLabels=columns,
        cellColours=cells_colours,
        cellLoc='center',
        colWidths=[0.2, 0.3, 0.4, 0.1, 0.2],
        bbox=[0, 0, 1.0, 1.0],
        loc='center'
    )

    table.auto_set_font_size(False)
    table.set_fontsize(14)

    axis.set_xticks([])
    axis.axis('off')
    axis.set_title('Tasks')

    return finish


def describe_logs(
        dag: int,
        axis,
        max_log_text: int = None,
        log_count: int = 5,
        col_withds: List[float] = None
):
    columns = ['Component', 'Level', 'Task', 'Time', 'Text']
    provider = LogProvider()
    logs = provider.last(log_count, dag=dag)

    res = []

    cells = []
    cells_colours = []

    for log, task_id in logs:
        component = to_snake(ComponentType(log.component).name)

        level = log.level
        level = 'debug' if level == 10 else 'info' \
            if level == 20 else 'warning' \
            if level == 30 else 'error'
        message = log.message
        if max_log_text:
            message = message[:max_log_text]
        log_cells = [
            component, level,
            str(task_id),
            log.time.strftime('%m.%d %H:%M:%S'), message
        ]

        cells.append(log_cells)

        level_color = 'lightblue' if level == 'info' else 'lightyellow' \
            if level == 'warning' else 'red' if level == 'error' else 'white'

        log_colours = ['white', level_color, 'white', 'white', 'white']
        cells_colours.append(log_colours)

        if level == 'error':
            res.append(log)

    col_withds = col_withds or [0.2, 0.1, 0.25, 0.2, 0.45]
    if len(cells) > 0:
        table = axis.table(
            cellText=cells,
            colLabels=columns,
            cellColours=cells_colours,
            cellLoc='center',
            colWidths=col_withds,
            bbox=[0, 0, 1, 1.0],
            loc='center'
        )

        table.auto_set_font_size(False)
        table.set_fontsize(14)

    axis.set_xticks([])
    axis.axis('off')
    axis.set_title('Logs')

    return res


def describe_dag(dag, axis):
    provider = DagProvider()
    graph = provider.graph(dag)

    status_colors = {
        'not_ran': '#808080',
        'queued': '#add8e6',
        'in_progress': '#bfff00',
        'failed': '#e83217',
        'stopped': '#cb88ea',
        'skipped': '#ffa500',
        'success': '#006400'
    }
    node_color = []
    edge_color = []

    G = nx.DiGraph()
    labels = dict()
    for n in graph['nodes']:
        G.add_node(n['id'])
        labels[n['id']] = n['id']
        node_color.append(status_colors[n['status']])

    edges = []
    for e in graph['edges']:
        G.add_edge(e['from'], e['to'])
        edges.append((e['from'], e['to']))
        edge_color.append(status_colors[e['status']])

    pos = nx.spring_layout(G, seed=0)
    nx.draw_networkx_nodes(
        G, pos, node_color=node_color, ax=axis, node_size=2000
    )
    nx.draw_networkx_labels(
        G,
        pos,
        labels,
        ax=axis,
        with_labels=True,
        font_color='orange',
        font_weight='bold',
        font_size=18
    )
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=edges,
        edge_color=edge_color,
        arrows=True,
        arrowsize=80,
        ax=axis
    )

    axis.set_xticks([])
    axis.axis('off')
    axis.set_title('Graph')


def describe_resources(computer: str, axis):
    provider = ComputerProvider()
    res = provider.get({})['data']
    res = [r for r in res if r['name'] == computer][0]
    usage = res['usage_history']
    x = [
        datetime.datetime.strptime(t, provider.datetime_format)
        for t in usage['time']
    ]

    for item in usage['mean']:
        if item['name'] == 'disk':
            continue

        axis.plot(x, item['value'], label=item['name'])

    axis.set_title('Resources')
    axis.set_ylabel('%')
    axis.legend(loc='lower left')


def describe_metrics(series, axis, last_n_epoch=None):
    for i in range(len(axis)):
        ax = axis[i]
        if i >= len(series):
            ax.axis('off')
            continue

        ax = axis[i]
        ax.axis('on')
        task_name, metric, groups = series[i]

        for group in groups:
            if last_n_epoch:
                group['epoch'] = group['epoch'][-last_n_epoch:]
                group['value'] = group['value'][-last_n_epoch:]

            ax.plot(group['epoch'], group['value'], label=group['name'])

        ax.set_title(f'{task_name}, {metric} score')
        ax.set_ylabel(metric, labelpad=20)
        ax.set_xlabel('epoch')
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.legend()


def describe_task_names(dag: int):
    pd.set_option('display.max_columns', None)
    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('max_colwidth', -1)

    provider = TaskProvider()
    tasks = provider.by_dag(dag)
    return pd.DataFrame([{'id': t.id, 'name': t.name} for t in tasks])


def describe(
        dag: int,
        metrics=None,
        last_n_epoch=None,
        computer: str = None,
        max_log_text: int = 45,
        fig_size=(12, 10),
        grid_spec: dict = None,
        log_count=5,
        log_col_widths: List[float] = None,
        wait=True,
        wait_interval=5,
        task_with_metric_count=0
):
    grid_spec = grid_spec or {}
    metrics = metrics or []

    series_count = task_with_metric_count * len(metrics)
    size = (4 + ceil(series_count / 2), 2)
    default_grid_spec = {
        'tasks': {
            'rowspan': 1,
            'colspan': 2,
            'loc': (0, 0)
        },
        'dag': {
            'rowspan': 1,
            'colspan': 2,
            'loc': (1, 0)
        },
        'logs': {
            'rowspan': 1,
            'colspan': 2,
            'loc': (2, 0)
        },
        'resources': {
            'rowspan': 1,
            'colspan': 2,
            'loc': (3, 0)
        },
        'size': size
    }

    loc = (4, 0)
    for i in range(series_count):
        default_grid_spec[i] = {'rowspan': 1, 'colspan': 1,
                                'loc': loc}
        if loc[1] == 1:
            loc = (loc[0] + 1, 0)
        else:
            loc = (loc[0], 1)

    default_grid_spec.update(grid_spec)
    grid_spec = default_grid_spec

    fig = plt.figure(figsize=fig_size)

    def grid_cell(spec: dict):
        return plt.subplot2grid(
            size,
            spec['loc'],
            colspan=spec['colspan'],
            rowspan=spec['rowspan'],
            fig=fig
        )

    while True:
        computer = computer or gethostname()

        task_axis = grid_cell(grid_spec['tasks'])
        dag_axis = grid_cell(grid_spec['dag'])
        resources_axis = grid_cell(grid_spec['resources'])
        logs_axis = grid_cell(grid_spec['logs'])

        finish = describe_tasks(dag, task_axis)
        describe_dag(dag, dag_axis)
        errors = describe_logs(
            dag,
            axis=logs_axis,
            max_log_text=max_log_text,
            log_count=log_count,
            col_withds=log_col_widths
        )
        describe_resources(computer=computer, axis=resources_axis)

        series_provider = ReportSeriesProvider()
        series = series_provider.by_dag(dag, metrics)

        metric_axis = [grid_cell(grid_spec[i]) for i, s in enumerate(series)]

        describe_metrics(series, last_n_epoch=last_n_epoch, axis=metric_axis)

        plt.tight_layout()

        display.clear_output(wait=True)

        for error in errors:
            print(error.time)
            print(error.message)

        display.display(fig)

        if not wait or finish:
            break

        time.sleep(wait_interval)

    plt.close(fig)


__all__ = ['describe']

if __name__ == '__main__':
    describe(dag=293, metrics=['loss', 'dice'], wait_interval=2)



import os
import logging
import sys
from logging.handlers import RotatingFileHandler

from mlcomp import LOG_FOLDER, LOG_NAME, FILE_LOG_LEVEL, DB_LOG_LEVEL, \
    CONSOLE_LOG_LEVEL
from mlcomp.db.core import Session
from mlcomp.db.providers import LogProvider
from mlcomp.db.models import Log
from mlcomp.utils.misc import now

ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))


class Formatter(logging.Formatter):
    def format(self, record):
        if not record.pathname.startswith(ROOT):
            return super().format(record)

        msg = str(record.msg)
        if record.args:
            try:
                msg = msg % record.args
            except Exception:
                pass
        record.message = msg
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self.formatMessage(record)
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + self.formatStack(record.stack_info)
        return s


class DbHandler(logging.Handler):
    """
    A handler class which writes logging records, appropriately formatted,
    to the database.
    """

    def __init__(self, session: Session):
        """
        Initialize the handler.
        """
        logging.Handler.__init__(self)
        self.provider = LogProvider(session)

    def emit(self, record):
        """
        Emit a record.
        """
        try:
            if not record.pathname.startswith(ROOT):
                return

            assert 1 <= len(record.args), \
                'Args weer not been provided for logging'
            assert len(record.args) <= 4, 'Too many args for logging'

            step = None
            task = None
            computer = None

            if len(record.args) == 1:
                component = record.args[0]
            elif len(record.args) == 2:
                component, computer = record.args
            elif len(record.args) == 3:
                component, computer, task = record.args
            else:
                component, computer, task, step = record.args

            if not isinstance(component, int):
                component = component.value

            module = os.path.relpath(record.pathname, ROOT). \
                replace(os.sep, '.').replace('.py', '')
            if record.funcName and record.funcName != '<module>':
                module = f'{module}:{record.funcName}'
            log = Log(
                message=record.msg[-16000:],
                time=now(),
                level=record.levelno,
                step=step,
                component=component,
                line=record.lineno,
                module=module,
                task=task,
                computer=computer
            )
            self.provider.add(log)
        except Exception:
            self.handleError(record)


def create_logger(session: Session, name: str, db=True, file=True,
                  console=True):
    logger = logging.Logger(name)
    logger.handlers = []

    if console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(CONSOLE_LOG_LEVEL)
        console_handler.stream = sys.stdout
        logger.handlers.append(console_handler)

    if file:
        file_path = os.path.join(LOG_FOLDER, f'{LOG_NAME}.txt')
        file_handler = RotatingFileHandler(file_path)
        file_handler.setLevel(FILE_LOG_LEVEL)
        file_handler.maxBytes = 10485760
        file_handler.backupCount = 1
        logger.handlers.append(file_handler)

    if db:
        handler = DbHandler(session)
        handler.setLevel(DB_LOG_LEVEL)
        logger.handlers.append(handler)

    for h in logger.handlers:
        fmt = '%(asctime)s.%(msecs)03d %(levelname)s' \
              ' %(module)s - %(funcName)s: %(message)s'
        datefmt = '%Y-%m-%d %H:%M:%S'
        if isinstance(h, DbHandler):
            fmt, datefmt = None, None
        h.formatter = Formatter(fmt=fmt, datefmt=datefmt)

    # ignore messages from some libraries
    class NoRunningFilter(logging.Filter):
        def filter(self, record):
            return 'ran tasks' not in str(record.msg)

    for k in logging.root.manager.loggerDict:
        if 'apscheduler' in k:
            logging.getLogger(k).setLevel(logging.ERROR)
        if 'mlcomp' in k:
            logging.getLogger(k).addFilter(NoRunningFilter())
        if 'serializer' in k:
            logging.getLogger(k).setLevel(logging.ERROR)

    return logger


__all__ = ['create_logger']



from collections import OrderedDict

from mlcomp.db.report_info.item import ReportLayoutItem


class ReportLayoutImgClassify(ReportLayoutItem):
    def __init__(
        self,
        name: str,
        confusion_matrix: bool,
    ):
        super().__init__(name)

        self.confusion_matrix = confusion_matrix

    @classmethod
    def from_dict(cls, name: str, value: OrderedDict):
        value.pop('type')
        confusion_matrix = value.pop('confusion_matrix', False)

        assert len(value) == 0, f'Unknown parameter in ' \
            f'report.img_classify={value.popitem()}'
        return cls(
            name,
            confusion_matrix=confusion_matrix
        )


__all__ = ['ReportLayoutImgClassify']



from typing import Tuple, List

import cv2
import numpy as np

from mlcomp.db.core import Session
from mlcomp.db.models import Report, ReportImg, Task, ReportTasks, ReportSeries
from mlcomp.db.providers import ReportProvider, ReportLayoutProvider, \
    TaskProvider, ReportImgProvider, ReportTasksProvider, \
    ReportSeriesProvider, DagProvider
from mlcomp.utils.img import resize_saving_ratio
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.utils.misc import now


class SegmentationReportBuilder:
    def __init__(
            self,
            session: Session,
            task: Task,
            layout: str,
            part: str = 'valid',
            name: str = 'img_segment',
            max_img_size: Tuple[int, int] = None,
            stack_type: str = 'vertical',
            main_metric: str = 'dice',
            plot_count: int = 0,
            colors: List[Tuple] = None
    ):
        self.session = session
        self.task = task
        self.layout = layout
        self.part = part
        self.name = name or 'img_segment'
        self.max_img_size = max_img_size
        self.stack_type = stack_type
        self.main_metric = main_metric
        self.colors = colors
        self.plot_count = plot_count

        self.dag_provider = DagProvider(session)
        self.report_provider = ReportProvider(session)
        self.layout_provider = ReportLayoutProvider(session)
        self.task_provider = TaskProvider(session)
        self.report_img_provider = ReportImgProvider(session)
        self.report_task_provider = ReportTasksProvider(session)
        self.report_series_provider = ReportSeriesProvider(session)

        self.project = self.task_provider.project(task.id).id
        self.layout = self.layout_provider.by_name(layout)
        self.layout_dict = yaml_load(self.layout.content)

        self.create_base()

    def create_base(self):
        report = Report(
            config=yaml_dump(self.layout_dict),
            time=now(),
            layout=self.layout.name,
            project=self.project,
            name=self.name
        )
        self.report_provider.add(report)
        self.report_task_provider.add(
            ReportTasks(report=report.id, task=self.task.id)
        )

        self.task.report = report.id
        self.task_provider.update()

    def encode_pred(self, mask: np.array):
        res = np.zeros((*mask.shape[1:], 3), dtype=np.uint8)
        for i, c in enumerate(mask):
            c = np.repeat(c[:, :, None], 3, axis=2)
            color = self.colors[i] if self.colors is not None else (
                255, 255, 255
            )
            res += (c * color).astype(np.uint8)

        return res

    def plot_mask(self, img: np.array, mask: np.array):
        if len(img.shape) == 2:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        img = img.astype(np.uint8)
        mask = mask.astype(np.uint8)

        for i, c in enumerate(mask):
            contours, _ = cv2.findContours(
                c, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE
            )
            color = self.colors[i] if self.colors else (0, 255, 0)
            for i in range(0, len(contours)):
                cv2.polylines(img, contours[i], True, color, 2)

        return img

    def process_scores(self, scores):
        for key, item in self.layout_dict['items'].items():
            item['name'] = key
            if item['type'] == 'series' and item['key'] in scores:
                series = ReportSeries(
                    name=item['name'],
                    value=scores[item['key']],
                    epoch=0,
                    time=now(),
                    task=self.task.id,
                    part='valid',
                    stage='stage1'
                )

                self.report_series_provider.add(series)

    def process_pred(self, imgs: np.array, preds: dict,
                     targets: np.array = None, attrs=None, scores=None):
        for key, item in self.layout_dict['items'].items():
            item['name'] = key
            if item['type'] != 'img_segment':
                continue

            report_imgs = []
            dag = self.dag_provider.by_id(self.task.dag)

            for i in range(len(imgs)):
                if self.plot_count <= 0:
                    break

                if targets is not None:
                    img = self.plot_mask(imgs[i], targets[i])
                else:
                    img = imgs[i]

                imgs_add = [img]
                for key, value in preds.items():
                    imgs_add.append(self.encode_pred(value[i]))

                for j in range(len(imgs_add)):
                    imgs_add[j] = resize_saving_ratio(imgs_add[j],
                                                      self.max_img_size)

                if self.stack_type == 'horizontal':
                    img = np.hstack(imgs_add)
                else:
                    img = np.vstack(imgs_add)

                attr = attrs[i] if attrs else {}

                score = None
                if targets is not None:
                    score = scores[self.main_metric][i]

                retval, buffer = cv2.imencode('.jpg', img)
                report_img = ReportImg(
                    group=item['name'],
                    epoch=0,
                    task=self.task.id,
                    img=buffer,
                    dag=self.task.dag,
                    part=self.part,
                    project=self.project,
                    score=score,
                    **attr
                )

                self.plot_count -= 1
                report_imgs.append(report_img)
                dag.img_size += report_img.size

            self.dag_provider.commit()
            self.report_img_provider.bulk_save_objects(report_imgs)


__all__ = ['SegmentationReportBuilder']



import os

from mlcomp.db.core import Session
from mlcomp.db.enums import DagType
from mlcomp.db.providers import DagProvider, ProjectProvider, ModelProvider
from mlcomp.db.models import Dag
from mlcomp.worker.storage import Storage


def dag_pipe(session: Session, config: dict, config_text: str = None):
    assert 'pipes' in config, 'pipe missed'

    info = config['info']

    storage = Storage(session)
    dag_provider = DagProvider(session)

    folder = os.getcwd()
    project = ProjectProvider(session).by_name(info['project']).id
    dag = dag_provider.add(
        Dag(
            config=config_text,
            project=project,
            name=info['name'],
            docker_img=info.get('docker_img'),
            type=DagType.Pipe.value
        )
    )
    storage.upload(folder, dag)

    # Change model dags which have the same name
    ModelProvider(session
                  ).change_dag(project=project, name=info['name'], to=dag.id)


__all__ = ['dag_pipe']



from .decoder import PSPDecoder
from ..base import EncoderDecoder
from ..encoders import get_encoder


class PSPNet(EncoderDecoder):
    """PSPNet_ is a fully convolution
     neural network for image semantic segmentation

    Args:
        encoder_name: name of classification model used as feature
                extractor to build segmentation model.
        encoder_weights: one of ``None`` (random initialization),
         ``imagenet`` (pre-training on ImageNet).
        psp_in_factor: one of 4, 8 and 16.
        Downsampling rate or in other words backbone depth
            to construct PSP module on it.
        psp_out_channels: number of filters in PSP block.
        psp_use_batchnorm: if ``True``, ``BatchNormalisation``
        layer between ``Conv2D`` and ``Activation`` layers
                is used.
        psp_aux_output: if ``True`` add auxiliary
        classification output for encoder training
        psp_dropout: spatial dropout rate between 0 and 1.
        classes: a number of classes for output
        (output shape - ``(batch, classes, h, w)``).
        activation: activation function
        used in ``.predict(x)`` method for inference.
            One of [``sigmoid``, ``softmax``, callable, None]

    Returns:
        ``torch.nn.Module``: **PSPNet**

    .. _PSPNet:
        https://arxiv.org/pdf/1612.01105.pdf
    """

    def __init__(
            self,
            encoder_name='resnet34',
            encoder_weights='imagenet',
            psp_in_factor=8,
            psp_out_channels=512,
            psp_use_batchnorm=True,
            psp_aux_output=False,
            classes=21,
            dropout=0.2,
            activation='softmax',
    ):
        encoder = get_encoder(encoder_name, encoder_weights=encoder_weights)

        decoder = PSPDecoder(
            encoder_channels=encoder.out_shapes,
            downsample_factor=psp_in_factor,
            psp_out_channels=psp_out_channels,
            final_channels=classes,
            dropout=dropout,
            aux_output=psp_aux_output,
            use_batchnorm=psp_use_batchnorm,
        )

        super().__init__(encoder, decoder, activation)

        self.name = 'psp-{}'.format(encoder_name)
